"""
Handles calculations associated with PSC Cost Recovery.
"""

from dataclasses import dataclass, field
import numpy as np

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts import psc_tools
from pyscnomics.econ.selection import (
    FluidType,
    InflationAppliedTo,
    TaxRegime,
    FTPTaxRegime,
    TaxSplitTypeCR,
    TaxType,
    DeprMethod,
    OtherRevenue
)


class SunkCostException(Exception):
    """Exception to raise for a misuse of Sunk Cost Method"""

    pass


class CostRecoveryException(Exception):
    """Exception to raise for a misuse of Cost Recovery attributes"""

    pass


@dataclass
class CostRecovery(BaseProject):
    """
    Dataclass that represents Cost Recovery (CR) contract.

    Parameters
    ----------
    oil_ftp_is_available: bool
        The availability of the Oil's First Tranche Petroleum (FTP).
    oil_ftp_is_shared: bool
        The condition whether the Oil FTP will be shared to the contractor or not.
    oil_ftp_portion: float
        The portion of the Oil FTP in fraction.
    gas_ftp_is_available: bool
        The availability of the Gas's First Tranche Petroleum (FTP).
    gas_ftp_is_shared: bool
        The condition whether the Gas FTP will be shared to the contractor or not.
    gas_ftp_portion: float
        The portion of the Gas FTP in fraction.
    tax_split_type: TaxSplitTypeCR
        The type of the contractor split.
    condition_dict: dict
        The condition dictionary of the split when the tax_split_type is not Conventional.
    indicator_rc_icp_sliding: np.ndarray
        The indicator used in the SLIDING_SCALE or R2C tax_split_type.
    oil_ctr_pretax_share: float | np.ndarray
        The Oil Contractor Pre-Tax Split or Oil Contractor Pre-Tax Split Share.
    gas_ctr_pretax_share: float | np.ndarray
        The Gas Contractor Pre-Tax Split or Oil Contractor Pre-Tax Split Share.
    oil_ic_rate: float
        The Oil's Investment Credit (IC) rate of the contract.
    gas_ic_rate: float
        The Gas's Investment Credit (IC) rate of the contract.
    ic_is_available: bool
        The condition whether if IC is available or not.
    oil_cr_cap_rate: float
        The Oil Cost Recovery cap rate.
    gas_cr_cap_rate: float
        The Gas Cost Recovery cap rate.
    oil_dmo_volume_portion: float
        The Oil's Domestic Market Obligation (DMO) volume portion.
    oil_dmo_fee_portion: float
        The Oil's DMO fee portion.
    oil_dmo_holiday_duration: int
        The duration of the Oil DMO Holiday in month unit.
    gas_dmo_volume_portion: float
        The Gas's Domestic Market Obligation (DMO) volume portion.
    gas_dmo_fee_portion: float
        The Gas's DMO fee portion.
    gas_dmo_holiday_duration: int
        The duration of the Gas's DMO Holiday in month unit.
    """
    oil_ftp_is_available: bool = field(default=True)
    oil_ftp_is_shared: bool = field(default=True)
    oil_ftp_portion: float = field(default=0.2)

    gas_ftp_is_available: bool = field(default=True)
    gas_ftp_is_shared: bool = field(default=True)
    gas_ftp_portion: float = field(default=0.2)

    tax_split_type: TaxSplitTypeCR = field(default=TaxSplitTypeCR.CONVENTIONAL)
    condition_dict: dict = field(default_factory=dict)
    indicator_rc_icp_sliding: np.ndarray = field(default=None)
    oil_ctr_pretax_share: float | np.ndarray = field(default=0.25)
    gas_ctr_pretax_share: float | np.ndarray = field(default=0.5)

    oil_ic_rate: float = field(default=0.0)
    gas_ic_rate: float = field(default=0.0)
    ic_is_available: bool = field(default=False)
    oil_cr_cap_rate: float = field(default=1.0)
    gas_cr_cap_rate: float = field(default=1.0)

    oil_dmo_volume_portion: float = field(default=0.25)
    oil_dmo_fee_portion: float = field(default=0.25)
    oil_dmo_holiday_duration: int = field(default=60)

    gas_dmo_volume_portion: float = field(default=1.0)
    gas_dmo_fee_portion: float = field(default=1.0)
    gas_dmo_holiday_duration: int = field(default=60)

    _oil_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _gas_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _oil_undepreciated_asset: float = field(default=None, init=False, repr=False)
    _gas_undepreciated_asset: float = field(default=None, init=False, repr=False)

    _oil_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _gas_non_capital: np.ndarray = field(default=None, init=False, repr=False)

    _split_ctr_oil: np.ndarray = field(default=None, init=False, repr=False)
    _split_ctr_gas: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ic: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ic_unrecovered: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ic_paid: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ic: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ic_unrecovered: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ic_paid: np.ndarray = field(default=None, init=False, repr=False)

    _oil_unrecovered_before_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _gas_unrecovered_before_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )

    _oil_cost_to_be_recovered: np.ndarray = field(default=None, init=False, repr=False)
    _gas_cost_to_be_recovered: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cost_recovery: np.ndarray = field(default=None, init=False, repr=False)
    _gas_cost_recovery: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ets_before_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ets_before_transfer: np.ndarray = field(default=None, init=False, repr=False)

    _transfer_to_oil: np.ndarray = field(default=None, init=False, repr=False)
    _transfer_to_gas: np.ndarray = field(default=None, init=False, repr=False)

    _oil_unrecovered_after_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _gas_unrecovered_after_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )

    _oil_cost_to_be_recovered_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _gas_cost_to_be_recovered_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )

    _oil_cost_recovery_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _gas_cost_recovery_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )

    _oil_ets_after_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ets_after_transfer: np.ndarray = field(default=None, init=False, repr=False)

    _oil_contractor_share: np.ndarray = field(default=None, init=False, repr=False)
    _oil_government_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_contractor_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_government_share: np.ndarray = field(default=None, init=False, repr=False)

    _oil_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _oil_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _gas_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _gas_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ddmo: np.ndarray = field(default=None, init=False, repr=False)

    _oil_taxable_income: np.ndarray = field(default=None, init=False, repr=False)
    _gas_taxable_income: np.ndarray = field(default=None, init=False, repr=False)

    _tax_rate_arr: np.ndarray = field(default=None, init=False, repr=False)

    _oil_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _gas_tax_payment: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)

    _oil_contractor_take: np.ndarray = field(default=None, init=False, repr=False)
    _gas_contractor_take: np.ndarray = field(default=None, init=False, repr=False)
    _oil_government_take: np.ndarray = field(default=None, init=False, repr=False)
    _gas_government_take: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _gas_cashflow: np.ndarray = field(default=None, init=False, repr=False)

    # Consolidated Attributes
    _consolidated_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_capital_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_intangible: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_opex: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_asr: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_undepreciated_asset: np.ndarray | float = field(
        default=None, init=False, repr=False
    )
    _consolidated_ftp: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ftp_ctr: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ftp_gov: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ic: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ic_unrecovered: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_ic_paid: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_unrecovered_before_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_cost_recovery_before_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_ets_before_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_unrecovered_after_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_cost_to_be_recovered_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_cost_recovery_after_tf: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_ets_after_transfer: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_contractor_share: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_government_share: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_taxable_income: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_tax_due: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_unpaid_tax_balance: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_net_share: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_contractor_take: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_government_take: np.ndarray = field(
        default=None, init=False, repr=False
    )
    _consolidated_cashflow: np.ndarray = field(default=None, init=False, repr=False)

    def _check_attributes(self):
        """
        Function to check the Cost Recovery input.
        -------

        """
        # Defining any attributes that in the form of fraction
        fraction_attributes = (
            ('oil_ftp_portion', self.oil_ftp_portion),
            ('gas_ftp_portion', self.gas_ftp_portion),
            ('oil_ctr_pretax_share', self.oil_ctr_pretax_share),
            ('gas_ctr_pretax_share', self.gas_ctr_pretax_share),
            ('oil_ic_rate', self.oil_ic_rate),
            ('gas_ic_rate', self.gas_ic_rate),
            ('oil_cr_cap_rate', self.oil_cr_cap_rate),
            ('gas_cr_cap_rate', self.gas_cr_cap_rate),
            ('oil_dmo_volume_portion', self.oil_dmo_volume_portion),
            ('oil_dmo_fee_portion', self.oil_dmo_fee_portion),
            ('gas_dmo_volume_portion', self.gas_dmo_volume_portion),
            ('gas_dmo_fee_portion', self.gas_dmo_fee_portion)
        )

        for attr_name, attr in fraction_attributes:
            if attr > 1.0 or attr < 0.0:
                range_type = 'exceeding 1.0' if attr > 1.0 else 'below 0.0'
                raise CostRecoveryException(
                    f"The {attr_name} value, {attr}, is {range_type}. "
                    f"The allowed range for this attribute is between 0.0 and 1.0."
                )

        discrete_attributes = (
            ('oil_dmo_holiday_duration', self.oil_dmo_holiday_duration),
            ('gas_dmo_holiday_duration', self.gas_dmo_holiday_duration)
        )

        for attr_name, attr in discrete_attributes:
            if attr < 0:
                raise CostRecoveryException(
                    f"The {attr_name} value, {attr}, is below 0. "
                    f"The minimum value for this attribute is 0."
                )

    def _get_rc_icp_pretax(self):
        """
        A Function to get the value of PreTax using Revenue over Cost (RC) or Indonesian Crude Price (ICP) sliding scale.

        Notes
        -------
        The structure of the dictionary is as the following:
        condition_dict =
        {
            'condition_1': {'bot_limit': 0,
                            'top_limit': 1,
                            'ctr_oil': 0.3137260,
                            'ctr_gas': 0.5498040,
                            'gov_oil': 0.686274,
                            'gov_gas': 0.45098,
                            'split_ctr_oil': 20,
                            'split_ctr_gas': 35,
                            'split_gov_oil': 80,
                            'split_gov_gas': 65},

            'condition_2': {'bot_limit': 1,
                            'top_limit': 1.2,
                            'ctr_oil': 0.27451,
                            'ctr_gas': 0.509804,
                            'gov_oil': 0.72549,
                            'gov_gas': 0.490196,
                            'split_ctr_oil': 17.5,
                            'split_ctr_gas': 32.5,
                            'split_gov_oil': 82.5,
                            'split_gov_gas': 67.5},

            'condition_3': {'bot_limit': 1.2,
                            'top_limit': 1.4,
                            'ctr_oil': 0.235295,
                            'ctr_gas': 0.470589,
                            'gov_oil': 0.764705,
                            'gov_gas': 0.529411,
                            'split_ctr_oil': 15,
                            'split_ctr_gas': 30,
                            'split_gov_oil': 85,
                            'split_gov_gas': 70}
        }

        Returns
        -------
        Modifying the following Cost Recovery attributes:
            self.oil_ctr_pretax_share
            self.gas_ctr_pretax_share

        """
        # Conditioning the given dictionary
        length_condition = len(self.condition_dict[list(self.condition_dict.keys())[0]])
        enum_value = list(range(1, length_condition + 1))
        condition_dict_new = {}
        for index, step in enumerate(enum_value):
            dict_isi = {'bot_limit': self.condition_dict['RC Bottom Limit'][index],
                        'top_limit': self.condition_dict['RC Top Limit'][index],
                        'ctr_oil': self.condition_dict['Pre Tax CTR Oil'][index],
                        'ctr_gas': self.condition_dict['Pre Tax CTR Gas'][index]}

            key_string = 'condition_' + str(step)
            condition_dict_new[key_string] = dict_isi

        self.condition_dict = condition_dict_new

        # Extract relevant values from the condition_dict dictionary
        bot_limits = np.array([self.condition_dict[c]['bot_limit'] for c in self.condition_dict])
        top_limits = np.array([self.condition_dict[c]['top_limit'] for c in self.condition_dict])
        ctr_oil_values = np.array([self.condition_dict[c]['ctr_oil'] for c in self.condition_dict])
        ctr_gas_values = np.array([self.condition_dict[c]['ctr_gas'] for c in self.condition_dict])

        # Create conditions using vectorized comparisons
        conditions = (bot_limits[:, np.newaxis] < self.indicator_rc_icp_sliding) & (
                    self.indicator_rc_icp_sliding <= top_limits[:, np.newaxis])

        # Calculate corresponding values using np.select()
        self.oil_ctr_pretax_share = np.select(conditions, ctr_oil_values[:, np.newaxis], default=np.nan)
        self.gas_ctr_pretax_share = np.select(conditions, ctr_gas_values[:, np.newaxis], default=np.nan)

    def _get_ftp(self):
        self._oil_ftp_ctr = np.zeros_like(self._oil_revenue)
        self._gas_ftp_ctr = np.zeros_like(self._gas_revenue)
        self._oil_ftp_gov = np.zeros_like(self._oil_revenue)
        self._gas_ftp_gov = np.zeros_like(self._gas_revenue)

        if self.oil_ftp_is_available:
            self._oil_ftp = self.oil_ftp_portion * self._oil_revenue
            if self.oil_ftp_is_shared:
                self._oil_ftp_ctr = self.oil_ctr_pretax_share * self._oil_ftp
            self._oil_ftp_gov = self._oil_ftp - self._oil_ftp_ctr

        if self.gas_ftp_is_available:
            self._gas_ftp = self.gas_ftp_portion * self._gas_revenue
            if self.gas_ftp_is_shared:
                self._gas_ftp_ctr = self.gas_ctr_pretax_share * self._gas_ftp
            self._gas_ftp_gov = self._gas_ftp - self._gas_ftp_ctr

    def _get_ic(
            self,
            revenue: np.ndarray,
            ftp: np.ndarray,
            cost_alloc: FluidType,
            ic_rate: float,
    ):

        # Condition where fluid is Oil:
        if cost_alloc == FluidType.GAS:
            capital_class = self._gas_capital_cost
        else:
            capital_class = self._oil_capital_cost

        # Applying the IC calculation to only true value
        ic_arr = np.where(np.asarray(capital_class.is_ic_applied) == True,
                          capital_class.cost * ic_rate,
                          0)

        ic_expenses = np.bincount(
            capital_class.expense_year - capital_class.start_year, weights=ic_arr
        )
        zeros = np.zeros(self.project_duration - len(ic_expenses))
        ic_total = np.concatenate((ic_expenses, zeros))

        # Unrec IC
        ic_unrecovered = np.cumsum(ic_total) - np.cumsum(revenue - ftp)
        ic_unrecovered = np.where(ic_unrecovered > 0, ic_unrecovered, 0)

        ic_unrec_next = np.roll(ic_unrecovered, 1)
        ic_unrec_next[0] = 0
        ic_paid = np.minimum(revenue - ftp, ic_total + ic_unrec_next)

        return ic_total, ic_unrecovered, ic_paid

    @staticmethod
    def _get_cost_recovery(
            revenue: np.ndarray,
            ftp: np.ndarray,
            ic: np.ndarray,
            depreciation: np.ndarray,
            non_capital: np.ndarray,
            cost_to_be_recovered: np.ndarray,
            cr_cap_rate: float,
    ) -> np.ndarray:
        """
        A function to get the array of cost recovery.

        Parameters
        ----------
        revenue: np.ndarray
            The array containing the revenue.
        ftp: np.ndarray
            The array containing the First Tranche Petroleum (FTP).
        ic: np.ndarray
            The array containing the Paid Investment Credit (IC).
        depreciation: np.ndarray
            The array containing the depreciated expenditures.
        non_capital: np.ndarray
            The array containing the non-capital expenditures.
            non_capital expenditures is consisting of:
            Intangible, Operating Expenditures (OPEX) and Abandonment Site and Restoration (ASR) Expenditures.
        cost_to_be_recovered: np.ndarray
            The array containing the cost to be recovered.
        cr_cap_rate: float
            The cost recovery cap rate.

        Returns
        -------
        out: np.ndarray
            The array of cost recovery.
        """

        return np.minimum(
            revenue - ftp - ic,
            ((depreciation + non_capital + cost_to_be_recovered) * cr_cap_rate),
        )
        # There is possibility for a bug in Cap rate
        # return (depreciation + non_capital + cost_to_be_recovered) * cr_cap_rate

    @staticmethod
    def _get_ets_before_transfer(
            revenue: np.ndarray,
            ftp_ctr: np.ndarray,
            ftp_gov: np.ndarray,
            ic: np.ndarray,
            cost_recovery: np.ndarray,
    ) -> np.ndarray:
        """
        A function to get the Equity To be Shared (ETS) before transfer.

        Parameters
        ----------
        revenue: np.ndarray
            The array containing the revenue.
        ftp_ctr: np.ndarray
            The array containing the Contractor's First Tranche Petroleum (FTP).
        ftp_gov: np.ndarray
            The array containing the Government's FTP.
        ic: np.ndarray
            The array containing the Paid Investment Credit (IC).
        cost_recovery
            The array containing the cost recovery.

        Returns
        -------
        out: np.ndarray
            The array of ETS before transfer.
        """

        result = revenue - (ftp_ctr + ftp_gov) - ic - cost_recovery
        tol = np.full_like(result, fill_value=1.0e-5)
        return np.where(result < tol, 0, result)

    @staticmethod
    def _get_ets_after_transfer(
            ets_before_tf: np.ndarray,
            revenue: np.ndarray,
            ftp_ctr: np.ndarray,
            ftp_gov: np.ndarray,
            ic: np.ndarray,
            cost_recovery: np.ndarray,
            transferred_in: np.ndarray,
            transferred_out: np.ndarray,
    ) -> np.ndarray:
        """
        A function to get the Equity To be Shared (ETS) before transfer.

        Parameters
        ----------
        revenue: np.ndarray
            The array containing the revenue.
        ftp_ctr: np.ndarray
            The array containing the Contractor's First Tranche Petroleum (FTP).
        ftp_gov: np.ndarray
            The array containing the Government's FTP.
        ic: np.ndarray
            The array containing the Paid Investment Credit (IC).
        cost_recovery: np.ndarray
            The array containing the cost recovery.
        transferred_in: np.ndarray
            The transferred cost into the cashflow.
        transferred_out: np.ndarray
            The transferred cost out from the cashflow.


        Returns
        -------
        out: np.ndarray
            The array of ETS before transfer.
        """
        ets_after_transfer = np.where(ets_before_tf >= 0,
                                      revenue
                                      - (ftp_ctr + ftp_gov)
                                      - ic
                                      - cost_recovery
                                      - transferred_out
                                      + transferred_in,
                                      0)

        tol = np.full_like(ets_after_transfer, fill_value=1.0e-5)
        return np.where(ets_after_transfer < tol, 0, ets_after_transfer)

    @staticmethod
    def _get_equity_share(ets, pretax_ctr) -> tuple:
        r"""
        Use to calculate equity share (ES).

        .. math::

            ES_{ctr \, (t)} &= pretax_ctr \times ETS_{(t)} \\
            ES_{gov \, (t)} &= (1 - pretax_ctr) \times ETS_{(t)}

        Parameters
        ----------
        ets : np.ndarray
            equity to be shared
        pretax_ctr: float
            pretax contractor

        Returns
        -------
        out: tup
            * ES_ctr : np.ndarray
                Equity share of contractor
            * ES_gov : np.ndarray
                Equity share of government
        """
        contractor_share = pretax_ctr * ets
        government_share = (1 - pretax_ctr) * ets

        return contractor_share, government_share

    @staticmethod
    def _get_tax_payment(
            ctr_share: np.ndarray,
            taxable_income: np.ndarray,
            tax_rate: np.ndarray,
            ftp_ctr: np.ndarray,
            unrec_cost: np.ndarray,
            ftp_tax_regime: FTPTaxRegime = FTPTaxRegime.PRE_PDJP_20_2017,
    ) -> np.ndarray:
        """
        The function to get the tax payment of the Cost Recovery contract based on the chosen regime.
        There are three existing regime, they are :
            - PRE_PDJP_20_2017
            - PDJP_20_2017
            - DIRECT_MODE

        Notes
        ----------
        [1] PRE_PDJP_20_2017
            In general, the tax will be applied after the contractor share is not zero.
            The taxable object that occurs at the time prior of the applied tax time will be accumulated and weighted
            on the time when the tax is applied.

        [2] PDJP_20_2017
            The tax will be applied when the cumulative First Tranche Petroleum (FTP) is greater than
            the unrecoverable cost

        [3] DIRECT_MODE
            The tax will be applied regardless the cumulative FTP, unrecoverable cost and Contractor Share condition.


        Parameters
        ----------
        ctr_share: np.ndarray
            The contractor equity to be split
        taxable_income: np.ndarray
            The contractor's Taxable Income (TI)
        tax_rate: float | np.ndarray
            The applied effective tax rate
        ftp_ctr: np.ndarray
            The contractor's First Tranche Petroleum (FTP)
        unrec_cost: np.ndarray
            The unrecoverable cost
        ftp_tax_regime: FTPTaxRegime
            The regime used to calculate the tax payment

        Returns
        -------
        ctr_tax: np.ndarray
            The contractor's tax payment

        """

        # Tax payment prior to PDJP 2017
        if ftp_tax_regime == FTPTaxRegime.PRE_PDJP_20_2017:
            applied_tax = np.where(ctr_share > 0, 1, 0)
            cum_ti = np.cumsum(taxable_income)

            applied_tax_prior = np.concatenate((np.zeros(1), applied_tax))[0:-1]

            ctr_tax = np.where(
                np.logical_and(applied_tax == 1, applied_tax_prior == 0),
                cum_ti * tax_rate,
                np.where(
                    np.logical_and(applied_tax == 1, applied_tax_prior == 1),
                    taxable_income * tax_rate,
                    0,
                ),
            )

        # Tax payment after PDJP 2017
        elif ftp_tax_regime == FTPTaxRegime.PDJP_20_2017:
            # Defining the array used in calculation of ftp tax payment
            ftp_cum_b2 = np.cumsum(ftp_ctr)
            ftp_prior_b3 = np.zeros_like(ftp_ctr, dtype=float)
            ftp_diff_b4 = np.zeros_like(ftp_ctr, dtype=float)
            ftp_considered_b5 = np.zeros_like(ftp_ctr, dtype=float)

            # Looping throughout the ftp array
            for i, value in enumerate(ftp_ctr):
                if i == 0:
                    ftp_prior_b3[i] = 0

                ftp_prior_b3[i] = ftp_prior_b3[i - 1] + ftp_considered_b5[i - 1]

                ftp_diff_b4[i] = np.where(ftp_cum_b2[i] > unrec_cost[i],
                                          ftp_cum_b2[i] - ftp_prior_b3[i],
                                          0)
                ftp_considered_b5[i] = np.where(ftp_diff_b4[i] > unrec_cost[i],
                                                ftp_diff_b4[i] - unrec_cost[i],
                                                0)

            # Calculating the ftp tax payment
            ftp_tax_paid = ftp_considered_b5 * tax_rate

            # Calculating the tax from taxable income
            ti_tax = (taxable_income - ftp_ctr) * tax_rate

            # Calculating ctr_tax
            ctr_tax = ftp_tax_paid + ti_tax

        else:
            ctr_tax = taxable_income * tax_rate

        return ctr_tax

    @staticmethod
    def _unpaid_and_tax_balance(tax_payment: np.ndarray, ets_ctr: np.ndarray):
        """
        The function to get the contractor's unpaid tax and tax payment.

        Parameters
        ----------
        tax_payment: np.ndarray
            The contractors tax payment
        ets_ctr: np.ndarray
            The contractor's share

        Returns
        -------
        unpaid_tax: np.ndarray
            The unpaid contractor's tax
        ctr_tax: np.ndarray
            The contractor's tax payment
        """
        unpaid_tax = np.zeros_like(ets_ctr, dtype=float)
        unpaid_tax[0] = 0
        ctr_tax = np.zeros_like(ets_ctr, dtype=float)

        for i in range(1, len(unpaid_tax)):
            unpaid_tax[i] = max(0, unpaid_tax[i - 1] + tax_payment[i] - ets_ctr[i])
            ctr_tax[i] = min(ets_ctr[i], unpaid_tax[i - 1] + tax_payment[i])

        return unpaid_tax, ctr_tax

    def _get_sunk_cost(self, sunk_cost_reference_year: int):
        oil_cost_raw = (
                self._oil_capital_expenditures
                + self._oil_non_capital
        )
        self._oil_sunk_cost = oil_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        gas_cost_raw = (
                self._gas_capital_expenditures
                + self._gas_non_capital
        )
        self._gas_sunk_cost = gas_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        if sunk_cost_reference_year == self.start_date.year:
            self._oil_sunk_cost = np.zeros(1)
            self._gas_sunk_cost = np.zeros(1)

    def run(
            self,
            sulfur_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            electricity_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_OIL_REVENUE,
            co2_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            is_dmo_end_weighted: bool = False,
            tax_regime: TaxRegime = TaxRegime.NAILED_DOWN,
            tax_rate: float | np.ndarray | None = None,
            ftp_tax_regime=FTPTaxRegime.PDJP_20_2017,
            sunk_cost_reference_year: int = None,
            depr_method: DeprMethod = DeprMethod.PSC_DB,
            decline_factor: float | int = 2,
            year_ref: int = None,
            tax_type: TaxType = TaxType.VAT,
            vat_rate: np.ndarray | float = 0.0,
            lbt_rate: np.ndarray | float = 0.0,
            inflation_rate: np.ndarray | float = 0.0,
            future_rate: float = 0.02,
            inflation_rate_applied_to: InflationAppliedTo | None = InflationAppliedTo.CAPEX,
            post_uu_22_year2001: bool = True
    ):

        # Configure Sunk Cost Reference Year
        if sunk_cost_reference_year is None:
            sunk_cost_reference_year = self.start_date.year

        if sunk_cost_reference_year > self.oil_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.oil_onstream_date}"
            )

        if sunk_cost_reference_year > self.gas_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.gas_onstream_date}"
            )

        if sunk_cost_reference_year < self.start_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is before the project start date: {self.start_date}"
            )

        if sunk_cost_reference_year > self.end_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the project end date: {self.end_date}"
            )

        # Configure year reference for expenditures
        if year_ref is None:
            year_ref = self.start_date.year

        # Get the WAP Price
        self._get_wap_price()

        # Calculate expenditures for every cost components
        self._get_expenditures(
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
            future_rate=future_rate,
            inflation_rate_applied_to=inflation_rate_applied_to
        )

        # Get The Other Revenue as the chosen selection
        self._get_other_revenue(sulfur_revenue=sulfur_revenue,
                                electricity_revenue=electricity_revenue,
                                co2_revenue=co2_revenue)

        # Calculate FTP
        self._get_ftp()

        # Defining the PreTax Split, whether using conventional PreTax or Sliding
        if self.tax_split_type is not TaxSplitTypeCR.CONVENTIONAL:
            self._get_rc_icp_pretax()

        # Depreciation (tangible cost)
        (
            self._oil_depreciation,
            self._oil_undepreciated_asset,
        ) = self._oil_capital_cost.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        (
            self._gas_depreciation,
            self._gas_undepreciated_asset,
        ) = self._gas_capital_cost.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        # Non-capital costs (intangible + opex + asr)
        self._oil_non_capital = (
                self._oil_intangible_expenditures
                + self._oil_opex_expenditures
                + self._oil_asr_expenditures
        )

        self._gas_non_capital = (
                self._gas_intangible_expenditures
                + self._gas_opex_expenditures
                + self._gas_asr_expenditures
        )

        # Filtering for only the cost that in the bracket of the project years
        self._oil_depreciation = self._oil_depreciation[
                                 : (self.end_date.year - self.start_date.year + 1)
                                 ]
        self._gas_depreciation = self._gas_depreciation[
                                 : (self.end_date.year - self.start_date.year + 1)
                                 ]

        # Get Sunk Cost
        self._get_sunk_cost(sunk_cost_reference_year)

        # Investment credit
        self._oil_ic, self._oil_ic_unrecovered, self._oil_ic_paid = self._get_ic(
            revenue=self._oil_revenue,
            ftp=self._oil_ftp,
            cost_alloc=FluidType.OIL,
            ic_rate=self.oil_ic_rate,
        )

        self._gas_ic, self._gas_ic_unrecovered, self._gas_ic_paid = self._get_ic(
            revenue=self._gas_revenue,
            ftp=self._gas_ftp,
            cost_alloc=FluidType.GAS,
            ic_rate=self.gas_ic_rate,
        )

        # Unrecovered cost before transfer/consolidation
        self._oil_unrecovered_before_transfer = psc_tools.get_unrecovered_cost(
            depreciation=self._oil_depreciation,
            non_capital=self._oil_non_capital,
            revenue=self._oil_revenue,
            ftp_ctr=self._oil_ftp_ctr,
            ftp_gov=self._oil_ftp_gov,
            ic=self._oil_ic_paid,
        )

        self._gas_unrecovered_before_transfer = psc_tools.get_unrecovered_cost(
            depreciation=self._gas_depreciation,
            non_capital=self._gas_non_capital,
            revenue=self._gas_revenue,
            ftp_ctr=self._gas_ftp_ctr,
            ftp_gov=self._gas_ftp_gov,
            ic=self._gas_ic_paid,
        )

        # Cost to be recovered
        self._oil_cost_to_be_recovered = psc_tools.get_cost_to_be_recovered(
            unrecovered_cost=self._oil_unrecovered_before_transfer,
        )

        self._gas_cost_to_be_recovered = psc_tools.get_cost_to_be_recovered(
            unrecovered_cost=self._gas_unrecovered_before_transfer,
        )

        # Cost recovery
        self._oil_cost_recovery = self._get_cost_recovery(
            revenue=self._oil_revenue,
            ftp=self._oil_ftp,
            ic=self._oil_ic_paid,
            depreciation=self._oil_depreciation,
            non_capital=self._oil_non_capital,
            cost_to_be_recovered=self._oil_cost_to_be_recovered,
            cr_cap_rate=self.oil_cr_cap_rate,
        )

        self._gas_cost_recovery = self._get_cost_recovery(
            revenue=self._gas_revenue,
            ftp=self._gas_ftp,
            ic=self._gas_ic_paid,
            depreciation=self._gas_depreciation,
            non_capital=self._gas_non_capital,
            cost_to_be_recovered=self._gas_cost_to_be_recovered,
            cr_cap_rate=self.gas_cr_cap_rate,
        )

        # ETS (Equity to be Split) before transfer/consolidation
        self._oil_ets_before_transfer = self._get_ets_before_transfer(
            revenue=self._oil_revenue,
            ftp_ctr=self._oil_ftp_ctr,
            ftp_gov=self._oil_ftp_gov,
            ic=self._oil_ic_paid,
            cost_recovery=self._oil_cost_recovery,
        )

        self._gas_ets_before_transfer = self._get_ets_before_transfer(
            revenue=self._gas_revenue,
            ftp_ctr=self._gas_ftp_ctr,
            ftp_gov=self._gas_ftp_gov,
            ic=self._gas_ic_paid,
            cost_recovery=self._gas_cost_recovery,
        )

        self._transfer_to_oil, self._transfer_to_gas = psc_tools.get_transfer(
            gas_unrecovered=self._gas_unrecovered_before_transfer,
            oil_unrecovered=self._oil_unrecovered_before_transfer,
            gas_ets_pretransfer=self._gas_ets_before_transfer,
            oil_ets_pretransfer=self._oil_ets_before_transfer,
        )

        self._oil_unrecovered_after_transfer = psc_tools.get_unrec_cost_after_tf(
            depreciation=self._oil_depreciation,
            non_capital=self._oil_non_capital,
            revenue=self._oil_revenue,
            ftp_ctr=self._oil_ftp_ctr,
            ftp_gov=self._oil_ftp_gov,
            ic=self._oil_ic_paid,
            transferred_cost_in=self._transfer_to_oil,
            transferred_cost_out=self._transfer_to_gas,
        )

        self._gas_unrecovered_after_transfer = psc_tools.get_unrec_cost_after_tf(
            depreciation=self._gas_depreciation,
            non_capital=self._gas_non_capital,
            revenue=self._gas_revenue,
            ftp_ctr=self._gas_ftp_ctr,
            ftp_gov=self._gas_ftp_gov,
            ic=self._gas_ic_paid,
            transferred_cost_in=self._transfer_to_gas,
            transferred_cost_out=self._transfer_to_oil,
        )

        # Adjust Transfer
        self._transfer_to_oil = psc_tools.transfer_treatment(
            unrecovered_prior_to_cost=self._oil_unrecovered_before_transfer,
            unrecovered_after_to_cost=self._oil_unrecovered_after_transfer,
            transfer_prior=self._transfer_to_oil)

        self._transfer_to_gas = psc_tools.transfer_treatment(
            unrecovered_prior_to_cost=self._gas_unrecovered_before_transfer,
            unrecovered_after_to_cost=self._gas_unrecovered_after_transfer,
            transfer_prior=self._transfer_to_gas)

        # Cost to be recovered after transfer
        self._oil_cost_to_be_recovered_after_tf = (
            psc_tools.get_cost_to_be_recovered_after_tf(
                unrecovered_cost=self._oil_unrecovered_after_transfer,
                transferred_cost=self._transfer_to_oil,
            )
        )

        self._gas_cost_to_be_recovered_after_tf = (
            psc_tools.get_cost_to_be_recovered_after_tf(
                unrecovered_cost=self._gas_unrecovered_after_transfer,
                transferred_cost=self._transfer_to_gas,
            )
        )

        # Cost recovery after transfer
        self._oil_cost_recovery_after_tf = (
                self._get_cost_recovery(
                    revenue=self._oil_revenue,
                    ftp=self._oil_ftp,
                    ic=self._oil_ic_paid,
                    depreciation=self._oil_depreciation,
                    non_capital=self._oil_non_capital,
                    cost_to_be_recovered=self._oil_cost_to_be_recovered_after_tf,
                    cr_cap_rate=self.oil_cr_cap_rate,
                )
                + self._transfer_to_oil
        )

        self._gas_cost_recovery_after_tf = (
                self._get_cost_recovery(
                    revenue=self._gas_revenue,
                    ftp=self._gas_ftp,
                    ic=self._gas_ic_paid,
                    depreciation=self._gas_depreciation,
                    non_capital=self._gas_non_capital,
                    cost_to_be_recovered=self._gas_cost_to_be_recovered_after_tf,
                    cr_cap_rate=self.gas_cr_cap_rate,
                )
                + self._transfer_to_gas
        )

        # ETS after Transfer
        self._oil_ets_after_transfer = self._get_ets_after_transfer(
            ets_before_tf=self._oil_ets_before_transfer,
            revenue=self._oil_revenue,
            ftp_ctr=self._oil_ftp_ctr,
            ftp_gov=self._oil_ftp_gov,
            ic=self._oil_ic_paid,
            cost_recovery=self._oil_cost_recovery_after_tf,
            transferred_in=self._transfer_to_oil,
            transferred_out=self._transfer_to_gas,
        )

        self._gas_ets_after_transfer = self._get_ets_after_transfer(
            ets_before_tf=self._gas_ets_before_transfer,
            revenue=self._gas_revenue,
            ftp_ctr=self._gas_ftp_ctr,
            ftp_gov=self._gas_ftp_gov,
            ic=self._gas_ic_paid,
            cost_recovery=self._gas_cost_recovery_after_tf,
            transferred_in=self._transfer_to_gas,
            transferred_out=self._transfer_to_oil,
        )

        # ES (Equity Share)
        self._oil_contractor_share, self._oil_government_share = self._get_equity_share(
            ets=self._oil_ets_after_transfer, pretax_ctr=self.oil_ctr_pretax_share
        )

        self._gas_contractor_share, self._gas_government_share = self._get_equity_share(
            ets=self._gas_ets_after_transfer, pretax_ctr=self.gas_ctr_pretax_share
        )

        # DMO
        self._oil_dmo_volume, self._oil_dmo_fee, self._oil_ddmo = psc_tools.get_dmo(
            onstream_date=self.oil_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.oil_dmo_holiday_duration,
            dmo_volume_portion=self.oil_dmo_volume_portion,
            dmo_fee_portion=self.oil_dmo_fee_portion,
            lifting=self._oil_lifting,
            price=self._oil_wap_price,
            ctr_pretax_share=self.oil_ctr_pretax_share,
            unrecovered_cost=self._oil_unrecovered_after_transfer,
            is_dmo_end_weighted=is_dmo_end_weighted,
            ets=self._oil_ets_after_transfer,
            ctr_ets=self._oil_contractor_share,
            ctr_ftp=self._oil_ftp_ctr,
            post_uu_22_year2001=post_uu_22_year2001,
        )

        self._gas_dmo_volume, self._gas_dmo_fee, self._gas_ddmo = psc_tools.get_dmo(
            onstream_date=self.gas_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.gas_dmo_holiday_duration,
            dmo_volume_portion=self.gas_dmo_volume_portion,
            dmo_fee_portion=self.gas_dmo_fee_portion,
            lifting=self._gas_lifting,
            price=self._gas_wap_price,
            ctr_pretax_share=self.gas_ctr_pretax_share,
            unrecovered_cost=self._gas_unrecovered_after_transfer,
            is_dmo_end_weighted=is_dmo_end_weighted,
            ets=self._gas_ets_after_transfer,
            ctr_ets=self._gas_contractor_share,
            ctr_ftp=self._gas_ftp_ctr,
            post_uu_22_year2001=post_uu_22_year2001,
        )

        # Adjusting DDMO for Pre PDJP
        if ftp_tax_regime is FTPTaxRegime.PRE_PDJP_20_2017:
            self._oil_ddmo = np.where(self._oil_contractor_share > 0,
                                      self._oil_ddmo,
                                      0)
            self._gas_ddmo = np.where(self._gas_contractor_share > 0,
                                      self._gas_ddmo,
                                      0)

        # Taxable income (also known as Net Contractor Share - NCS)
        self._oil_taxable_income = (
                self._oil_ftp_ctr
                + self._oil_contractor_share
                + self._oil_ic_paid
                - self._oil_ddmo
        )
        self._gas_taxable_income = (
                self._gas_ftp_ctr
                + self._gas_contractor_share
                + self._gas_ic_paid
                - self._gas_ddmo
        )

        # Tax Payment
        # Generating Tax array if tax_rate argument is a single value not array
        if isinstance(tax_rate, float):
            self._tax_rate_arr = np.full_like(self.project_years, tax_rate, dtype=float)

        # Generating Tax array based on the tax regime if tax_rate argument is None
        elif tax_rate is None:
            self._tax_rate_arr = self._get_tax_by_regime(tax_regime=tax_regime)
        elif isinstance(tax_rate, np.ndarray):
            self._tax_rate_arr = tax_rate
        else:
            self._tax_rate_arr = np.full_like(self.project_years, fill_value=0.0, dtype=float)

        self._oil_tax_payment = self._get_tax_payment(
            ctr_share=self._oil_contractor_share,
            taxable_income=self._oil_taxable_income,
            tax_rate=self._tax_rate_arr,
            ftp_ctr=self._oil_ftp_ctr,
            unrec_cost=self._oil_unrecovered_after_transfer,
            ftp_tax_regime=ftp_tax_regime,
        )

        self._gas_tax_payment = self._get_tax_payment(
            ctr_share=self._gas_contractor_share,
            taxable_income=self._gas_taxable_income,
            tax_rate=self._tax_rate_arr,
            ftp_ctr=self._gas_ftp_ctr,
            unrec_cost=self._gas_unrecovered_after_transfer,
            ftp_tax_regime=ftp_tax_regime,
        )

        # Contractor Share
        self._oil_ctr_net_share = self._oil_taxable_income - self._oil_tax_payment
        self._gas_ctr_net_share = self._gas_taxable_income - self._gas_tax_payment

        # Contractor Take by Fluid
        self._oil_contractor_take = (
                self._oil_taxable_income
                - self._oil_tax_payment
                + self._oil_cost_recovery_after_tf
        )

        self._gas_contractor_take = (
                self._gas_taxable_income
                - self._gas_tax_payment
                + self._gas_cost_recovery_after_tf
        )

        # Contractor CashFlow
        self._oil_cashflow = self._oil_contractor_take - (
                self._oil_capital_expenditures + self._oil_non_capital
        )
        self._gas_cashflow = self._gas_contractor_take - (
                self._gas_capital_expenditures + self._gas_non_capital
        )

        # Government Take by Fluid
        self._oil_government_take = (
                self._oil_ftp_gov
                + self._oil_government_share
                + self._oil_tax_payment
                + self._oil_ddmo
        )

        self._gas_government_take = (
                self._gas_ftp_gov
                + self._gas_government_share
                + self._gas_tax_payment
                + self._gas_ddmo
        )

        # Consolidated attributes
        self._consolidated_revenue = self._oil_revenue + self._gas_revenue
        self._consolidated_capital_cost = (
                self._oil_capital_expenditures + self._gas_capital_expenditures
        )
        self._consolidated_intangible = (
                self._oil_intangible_expenditures + self._gas_intangible_expenditures
        )
        self._consolidated_sunk_cost = self._oil_sunk_cost + self._gas_sunk_cost
        self._consolidated_opex = (
                self._oil_opex_expenditures + self._gas_opex_expenditures
        )
        self._consolidated_asr = (
                self._oil_asr_expenditures + self._gas_asr_expenditures
        )
        self._consolidated_non_capital = self._oil_non_capital + self._gas_non_capital
        self._consolidated_depreciation = (
                self._oil_depreciation + self._gas_depreciation
        )

        self._consolidated_undepreciated_asset = (
                self._oil_undepreciated_asset + self._gas_undepreciated_asset
        )
        self._consolidated_ftp = self._oil_ftp + self._gas_ftp
        self._consolidated_ftp_ctr = self._oil_ftp_ctr + self._gas_ftp_ctr
        self._consolidated_ftp_gov = self._oil_ftp_gov + self._gas_ftp_gov
        self._consolidated_ic = self._oil_ic + self._gas_ic
        self._consolidated_ic_unrecovered = (
                self._oil_ic_unrecovered + self._gas_ic_unrecovered
        )
        self._consolidated_ic_paid = self._oil_ic_paid + self._gas_ic_paid
        self._consolidated_unrecovered_before_transfer = (
                self._oil_unrecovered_before_transfer
                + self._gas_unrecovered_before_transfer
        )
        self._consolidated_cost_recovery_before_transfer = (
                self._oil_cost_recovery + self._gas_cost_recovery
        )
        self._consolidated_ets_before_transfer = (
                self._oil_ets_before_transfer + self._gas_ets_before_transfer
        )
        self._consolidated_unrecovered_after_transfer = (
                self._oil_unrecovered_after_transfer + self._gas_unrecovered_after_transfer
        )
        self._consolidated_cost_to_be_recovered_after_tf = (
                self._oil_cost_to_be_recovered_after_tf
                + self._gas_cost_to_be_recovered_after_tf
        )
        self._consolidated_cost_recovery_after_tf = (
                self._oil_cost_recovery_after_tf + self._gas_cost_recovery_after_tf
        )
        self._consolidated_ets_after_transfer = (
                self._oil_ets_after_transfer + self._gas_ets_after_transfer
        )
        self._consolidated_contractor_share = (
                self._oil_contractor_share + self._gas_contractor_share
        )
        self._consolidated_government_share = (
                self._oil_government_share + self._gas_government_share
        )
        self._consolidated_dmo_volume = self._oil_dmo_volume + self._gas_dmo_volume
        self._consolidated_dmo_fee = self._oil_dmo_fee + self._gas_dmo_fee
        self._consolidated_ddmo = self._oil_ddmo + self._gas_ddmo
        self._consolidated_taxable_income = (
                self._oil_taxable_income + self._gas_taxable_income
        )

        # Calculating the consolidated tax based on the ftp tax payment regime
        if ftp_tax_regime is FTPTaxRegime.PDJP_20_2017:
            self._consolidated_tax_payment = self._oil_tax_payment + self._gas_tax_payment

        elif ftp_tax_regime is FTPTaxRegime.PRE_PDJP_20_2017:
            self._consolidated_tax_due = self._get_tax_payment(
                ctr_share=self._consolidated_contractor_share,
                taxable_income=self._consolidated_taxable_income,
                tax_rate=self._tax_rate_arr,
                ftp_ctr=self._consolidated_ftp_ctr,
                unrec_cost=self._consolidated_unrecovered_after_transfer,
                ftp_tax_regime=ftp_tax_regime,
            )
            (
                self._consolidated_unpaid_tax_balance,
                self._consolidated_tax_payment,
            ) = self._unpaid_and_tax_balance(
                tax_payment=self._consolidated_tax_due,
                ets_ctr=self._consolidated_contractor_share,
            )

        else:
            self._consolidated_tax_payment = self._oil_tax_payment + self._gas_tax_payment

        self._consolidated_ctr_net_share = (
                self._consolidated_taxable_income - self._consolidated_tax_payment
        )

        self._consolidated_contractor_take = (
                self._consolidated_taxable_income
                - self._consolidated_tax_payment
                + self._consolidated_cost_recovery_after_tf
        )

        self._consolidated_government_take = (
                self._consolidated_ftp_gov
                + self._consolidated_government_share
                + self._consolidated_tax_payment
                + self._consolidated_ddmo
        )

        self._consolidated_cashflow = self._consolidated_contractor_take - (
                self._consolidated_capital_cost + self._consolidated_non_capital
        )



import numpy as np
import dateutils
from datetime import date

from pyscnomics.econ.revenue import Lifting


def get_unrecovered_cost(depreciation: np.ndarray,
                         non_capital: np.ndarray,
                         revenue: np.ndarray,
                         ftp_ctr: np.ndarray,
                         ftp_gov: np.ndarray,
                         ic: np.ndarray) -> np.ndarray:
    """
    A function to get the array of unrecovered cost.

    Parameters
    ----------
    depreciation: np.ndarray
        The array containing the depreciated expenditures.
    non_capital: : np.ndarray
        The array containing the non-capital expenditures.
        non_capital expenditures is consisting of:
        Intangible, Operating Expenditures (OPEX) and Abandonment Site and Restoration (ASR) Expenditures.
    revenue: np.ndarray
        The array containing the revenue.
    ftp_ctr: np.ndarray
        The array containing the Contractor's First Tranche Petroleum (FTP).
    ftp_gov: np.ndarray
        The array containing the Government's FTP.
    ic: np.ndarray
        The array containing the Paid Investment Credit (IC).
    Returns
    -------
    out: np.ndarray
        The array of Unrecovered Cost.
    """

    unrecovered_cost = np.cumsum(depreciation + non_capital) - np.cumsum(
        revenue - (ftp_ctr + ftp_gov) - ic
    )

    unrecovered_cost = np.where(unrecovered_cost >= 0, unrecovered_cost, 0)

    # Condition where there is no revenue but still there is depreciation + non-capital
    left_cost = np.where(np.logical_and((revenue - ftp_ctr - ftp_gov - ic) < depreciation + non_capital,
                                        unrecovered_cost == 0),
                         (depreciation + non_capital) - (revenue - ftp_ctr - ftp_gov - ic), 0)

    unrecovered_cost_final = unrecovered_cost + left_cost

    # Adding the trailing unrecoverable cost into the array
    if np.sum(revenue) == 0:
        pass
    else:
        if np.sum(left_cost) == 0:
            pass
        else:
            last_non_zero_indices = max(np.nonzero(revenue)[0])
            trailing_cost = np.cumsum(left_cost[last_non_zero_indices:])

            unrecovered_cost_final[last_non_zero_indices:] = trailing_cost
    return unrecovered_cost_final


def get_cost_to_be_recovered(unrecovered_cost: np.ndarray) -> np.ndarray:
    """
    A function to get the array of cost to be recovered.

    Parameters
    ----------
    unrecovered_cost: np.ndarray
        The array containing the unrecovered cost.

    Returns
    -------
    out: np.ndarray
        The array of cost to be recovered.
    """
    ctr = np.concatenate((np.zeros(1), -np.diff(unrecovered_cost)))
    return np.where(ctr > 0, ctr, 0)


def get_cost_to_be_recovered_after_tf(unrecovered_cost: np.ndarray,
                                      transferred_cost: np.ndarray) -> np.ndarray:
    """
    A function to get the array of cost to be recovered.

    Parameters
    ----------
    unrecovered_cost: np.ndarray
        The array containing the unrecovered cost.
    transferred_cost

    Returns
    -------
    out: np.ndarray
        The array of cost to be recovered.
    """
    ctr = np.concatenate((np.zeros(1), -np.diff(unrecovered_cost - np.cumsum(transferred_cost))))
    result = np.where(ctr > 0, ctr, 0)

    return result


def get_transfer(gas_unrecovered: np.ndarray,
                 oil_unrecovered: np.ndarray,
                 gas_ets_pretransfer: np.ndarray,
                 oil_ets_pretransfer: np.ndarray) -> tuple:
    """
    A function to get the transferred cost between oil and gas.

    Parameters
    ----------
    gas_unrecovered: np.ndarray
        The array containing the unrecovered cost from gas.
    oil_unrecovered: np.ndarray
        The array containing the unrecovered cost from oil.
    gas_ets_pretransfer: np.ndarray
        The array containing the gas's Equity To be Split (ETS) before transfer.
    oil_ets_pretransfer: np.ndarray
        The array containing the oil's Equity To be Split (ETS) before transfer.

    Returns
    -------
    out: tuple
        trf2oil: np.ndarray
            The transferred cost from gas to oil.
        trf2gas: np.ndarray
            The transferred cost from oil to gas.
    """

    trf2oil = np.zeros_like(oil_unrecovered)
    trf2gas = np.zeros_like(gas_unrecovered)

    # Transfer to oil
    combined_condition_oil = np.logical_and(
        np.greater(oil_unrecovered, 0), np.equal(gas_unrecovered, 0)
    )

    oil_indices = np.argwhere(combined_condition_oil)

    if np.size(oil_indices) > 0:
        trf2oil[oil_indices] = np.minimum(
            gas_ets_pretransfer[oil_indices], oil_unrecovered[oil_indices]
        )

    # Transfer to gas
    combined_condition_gas = np.logical_and(
        np.equal(oil_unrecovered, 0), np.greater(gas_unrecovered, 0)
    )

    indices_gas = np.argwhere(combined_condition_gas)

    if np.size(indices_gas) > 0:
        trf2gas[indices_gas] = np.minimum(
            oil_ets_pretransfer[indices_gas], gas_unrecovered[indices_gas]
        )

    return trf2oil, trf2gas


# def get_unrec_cost_after_tf(depreciation,
#                             non_capital,
#                             revenue,
#                             ftp_ctr,
#                             ftp_gov,
#                             ic,
#                             transferred_cost):
#     unrecovered_cost = np.cumsum(depreciation + non_capital) - np.cumsum(
#         revenue - (ftp_ctr + ftp_gov) - ic
#     )
#
#     unrecovered_cost = np.where(unrecovered_cost >= 0, unrecovered_cost - np.cumsum(transferred_cost), 0)
#
#     # Condition where there is no revenue but still there is depreciation + non-capital
#     left_cost = np.where(np.logical_and((revenue - ftp_ctr - ftp_gov - ic) < depreciation + non_capital,
#                                         unrecovered_cost == 0),
#                          (depreciation + non_capital) - (revenue - ftp_ctr - ftp_gov - ic) - transferred_cost, 0)
#     unrecovered_cost_final = unrecovered_cost + np.cumsum(left_cost)
#     return unrecovered_cost_final

def get_unrec_cost_after_tf(depreciation,
                            non_capital,
                            revenue,
                            ftp_ctr,
                            ftp_gov,
                            ic,
                            transferred_cost_in,
                            transferred_cost_out):
    unrecovered_cost = np.cumsum(depreciation + non_capital) - np.cumsum(
        revenue - (ftp_ctr + ftp_gov) - ic
    )

    unrecovered_cost = np.where(unrecovered_cost >= 0, unrecovered_cost - np.cumsum(transferred_cost_in) + np.cumsum(transferred_cost_out), 0)
    unrecovered_cost = np.where(unrecovered_cost >= 0, unrecovered_cost, 0)

    # Condition where there is no revenue but still there is depreciation + non-capital
    left_cost = np.where(np.logical_and((revenue - ftp_ctr - ftp_gov - ic) < depreciation + non_capital,
                                        unrecovered_cost == 0),
                         (depreciation + non_capital) - (revenue - ftp_ctr - ftp_gov - ic) - transferred_cost_in + transferred_cost_out, 0)
    left_cost = np.where(left_cost >= 0, left_cost, 0)

    unrecovered_cost_final = unrecovered_cost + left_cost

    # Adding the trailing unrecoverable cost into the array
    if np.sum(revenue) == 0:
        pass
    else:
        if np.sum(left_cost) == 0:
            pass
        else:
            last_non_zero_indices = max(np.nonzero(revenue)[0])
            trailing_cost = np.cumsum(left_cost[last_non_zero_indices:])

            unrecovered_cost_final[last_non_zero_indices:] = trailing_cost
    return unrecovered_cost_final


def get_ets_after_transfer(ets_before_transfer: np.ndarray,
                           trfto: np.ndarray,
                           unrecovered_after_transfer: np.ndarray):
    """
    A function to get the Equity To be Split (ETS) after transfer.

    Parameters
    ----------
    ets_before_transfer: np.ndarray
        The array containing the ETS before transfer.
    trfto
        The array containing the transferred cost.
    unrecovered_after_transfer
        The array containing the unrecovered cost after transfer.

    Returns
    -------
    ets_after_transfer: np.ndarray
        The array of ETS after transfer.
    """

    ets_after_transfer = np.zeros_like(ets_before_transfer)

    indices = np.equal(unrecovered_after_transfer, 0)

    if np.size(indices) > 0:
        ets_after_transfer[indices] = ets_before_transfer[indices] + trfto[indices]

    return ets_after_transfer


# def get_ets_after_transfer(ets_before_transfer: np.ndarray,
#                            trfto: np.ndarray,
#                            trffrom: np.ndarray,
#                            unrecovered_after_transfer: np.ndarray):
#     """
#     A function to get the Equity To be Split (ETS) after transfer.
#
#     Parameters
#     ----------
#     ets_before_transfer: np.ndarray
#         The array containing the ETS before transfer.
#     trfto
#         The array containing the transferred cost.
#     unrecovered_after_transfer
#         The array containing the unrecovered cost after transfer.
#
#     Returns
#     -------
#     ets_after_transfer: np.ndarray
#         The array of ETS after transfer.
#     """
#
#     ets_after_transfer = np.zeros_like(ets_before_transfer)
#
#     indices = np.equal(unrecovered_after_transfer, 0)
#
#     if np.size(indices) > 0:
#         ets_after_transfer[indices] = ets_before_transfer[indices] - trfto[indices] + trffrom[indices]
#
#     return ets_after_transfer


def get_dmo(onstream_date: date,
            start_date: date,
            project_years: np.ndarray,
            dmo_holiday_duration: int,
            dmo_volume_portion: float,
            dmo_fee_portion: float,
            lifting: Lifting,
            price: np.ndarray,
            ctr_pretax_share: float,
            unrecovered_cost: np.ndarray,
            is_dmo_end_weighted: bool,
            ets: np.ndarray | None = None,
            ctr_ets: np.ndarray | None = None,
            ctr_ftp: np.ndarray | None = None,
            post_uu_22_year2001: bool = True,) -> tuple:
    """
    A function to get the array of Domestic Market Obligation (DMO).

    Parameters
    ----------
    onstream_date: date
        The onstream date.
    start_date: date
        The start date of the contract.
    project_years: np.ndarray
        The array containing the contract years from the beginning until the end of the contract.
    dmo_holiday_duration: int
        The duration of the DMO holiday.
    dmo_volume_portion: float
        The DMO volume portion.
    dmo_fee_portion: float
        The DMO fee portion.
    lifting: Lifting
        The Lifting class of the produced fluid.
    price: np.ndarray
        The Weighted Average Price (WAP) price of the produced fluid.
    ctr_pretax_share: float
        The Contractor PreTax.
    unrecovered_cost: np.ndarray
        The array containing the Unrecovered Cost.
    is_dmo_end_weighted: bool
        The condition of whether DMO is weighted or not.
    ets: np.ndarray | None
        The array of Equity to be Split (ETS).
    ctr_ets: np.ndarray | None
        The array contractor Equity to be Split (ETS).
    ctr_ftp: np.ndarray | None
        The array contractor First Tranche Petroleum (FTP).
    post_uu_22_year2001: bool
        The condition when the DMO is calculated using the DMO scheme pre or post UU No.21 Year 2001.

    Returns
    -------
    out: tuple
        dmo_volume: np.ndarray
            The array of DMO Volume.
        dmo_fee: np.ndarray
            The array of DMO Fee Volume.
        ddmo: np.ndarray
            The array of Difference DMO Fee.
    """

    # DMO end date
    dmo_end_date = onstream_date + dateutils.relativedelta(
        months=dmo_holiday_duration
    )

    # Identify position of dmo start year in project years array
    dmo_indices = onstream_date.year - start_date.year

    # Calculate DMO volume
    dmo_holiday = np.where(project_years >= dmo_end_date.year, False, True)

    # Condition when the DMO scheme is using pre UU No.22 Year 2001
    if post_uu_22_year2001 is False:
        dmo_volume = np.where(ets > 0,
                           np.minimum((dmo_volume_portion * lifting.get_lifting_rate_arr() * ctr_pretax_share),
                                      np.divide(ctr_ets, lifting.get_price_arr(), where=lifting.get_price_arr() != 0) + np.divide(ctr_ftp, lifting.get_price_arr(), where=lifting.get_price_arr() != 0)),
                           0)
    else:
        dmo_volume = dmo_volume_portion * lifting.get_lifting_rate_arr() * ctr_pretax_share

    dmo_fee = np.where(np.logical_and(unrecovered_cost == 0, ~dmo_holiday),
                       dmo_fee_portion * price * dmo_volume,
                       dmo_volume * price)

    # Weighted dmo fee condition if the period of dmo is ended in the middle of the year
    if unrecovered_cost[dmo_indices] > 0 and is_dmo_end_weighted and dmo_holiday[dmo_indices] == False:
        dmo_fee[dmo_indices] = (
                    dmo_end_date.month / 12 * price[dmo_indices] * dmo_volume[dmo_indices] +
                    (1 - dmo_end_date.month / 12) * dmo_volume[dmo_indices] *
                    dmo_fee_portion * price[dmo_indices])

    # Calculate Net DMO
    ddmo = (dmo_volume * price) - dmo_fee

    return dmo_volume, dmo_fee, ddmo


def get_dmo_gross_split(
        onstream_date: date,
        start_date: date,
        project_years: np.ndarray,
        dmo_holiday_duration: int,
        dmo_volume_portion: float,
        dmo_fee_portion: float,
        price: np.ndarray,
        unrecovered_cost: np.ndarray,
        is_dmo_end_weighted: bool,
        net_operating_profit: np.ndarray,
        contractor_share: np.ndarray) -> tuple:
    """
    A function to get the array of Domestic Market Obligation (DMO) in Gross Split contract.

    Parameters
    ----------
    onstream_date: date
        The onstream date.
    start_date: date
        The start date of the contract.
    project_years: np.ndarray
        The array containing the contract years from the beginning until the end of the contract.
    dmo_holiday_duration: int
        The duration of the DMO holiday.
    dmo_volume_portion: float
        The DMO volume portion.
    dmo_fee_portion: float
        The DMO fee portion.
    price: np.ndarray
        The Weighted Average Price (WAP) price of the produced fluid.
    unrecovered_cost: np.ndarray
        The array containing the Unrecovered Cost.
    is_dmo_end_weighted: bool
        The condition of whether DMO is weighted or not.
    net_operating_profit: np.ndarray
        The Net Operating Profit of the contractor.
    contractor_share: np.ndarray
        The contractor share.

    Returns
    -------
    out: tuple
        dmo_volume: np.ndarray
            The array of DMO Volume.
        dmo_fee: np.ndarray
            The array of DMO Fee Volume.
        ddmo: np.ndarray
            The array of Difference DMO Fee.

    Notes
    -------
    The difference with Cost Recovery DMO is that the DMO in Gross Split is using the following formula:
    DMO Gross Split = IF(net_operating_profit > 0; MIN(DMO Volume Portion * Contractor Share; net_operating_profit); 0)
    """

    # DMO end date
    dmo_end_date = onstream_date + dateutils.relativedelta(
        months=dmo_holiday_duration
    )

    # Identify position of dmo start year in project years array
    dmo_indices = onstream_date.year - start_date.year

    # Calculate DMO volume
    dmo_holiday = np.where(project_years >= dmo_end_date.year, False, True)

    # Calculate the Gross Split DMO
    dmo_volume = np.where(net_operating_profit > 0,
                          np.minimum(dmo_volume_portion * contractor_share, net_operating_profit),
                          0)
    dmo_volume = np.divide(dmo_volume, price, out=np.zeros_like(dmo_volume, dtype=float), where=price != 0)

    dmo_fee = np.where(np.logical_and(unrecovered_cost == 0, ~dmo_holiday),
                       dmo_fee_portion * price * dmo_volume,
                       dmo_volume * price)

    # Weighted dmo fee condition if the period of dmo is ended in the middle of the year
    if unrecovered_cost[dmo_indices] > 0 and is_dmo_end_weighted and dmo_holiday[dmo_indices] == False:
        dmo_fee[dmo_indices] = (
                    dmo_end_date.month / 12 * price[dmo_indices] * dmo_volume[dmo_indices] +
                    (1 - dmo_end_date.month / 12) * dmo_volume[dmo_indices] *
                    dmo_fee_portion * price[dmo_indices])

    # Calculate Net DMO
    ddmo = (dmo_volume * price) - dmo_fee

    return dmo_volume, dmo_fee, ddmo


def transfer_treatment(unrecovered_prior_to_cost: np.ndarray,
                       unrecovered_after_to_cost: np.ndarray,
                       transfer_prior: np.ndarray):
    # Get the cumulative transferred cost
    transfer_cum = np.cumsum(transfer_prior)

    # Get the unrec - cumulative transferred_cost
    difference = unrecovered_prior_to_cost - transfer_cum

    abs_of_diff = np.absolute(np.where(unrecovered_after_to_cost == 0, difference, 0))

    diff_to_prior = np.concatenate((np.zeros(1), np.diff(abs_of_diff)))

    positive_diff_to_prior = np.where(diff_to_prior < 0,
                                      0,
                                      diff_to_prior)
    transfer_adjusted = transfer_prior - positive_diff_to_prior
    transfer_final = np.where(transfer_adjusted < 0, 0, transfer_adjusted)
    #
    # df = pd.DataFrame()
    # df['Check'] = transfer_final
    # df.loc['Column_Total'] = df.sum(numeric_only=True, axis=0)
    # print(df)
    # input()

    return transfer_final




from dataclasses import dataclass, field
import numpy as np
from datetime import date

from pyscnomics.econ.revenue import Lifting
from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR
from pyscnomics.econ.selection import FluidType
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts import psc_tools


def adjust_rows(original_value: float | np.ndarray,
                first_contract: bool,
                prior_rows: np.ndarray,
                post_rows: np.ndarray, ):
    """
    Function to adjust an array based on the given condition, tailored for transition contract.
    Parameters
    ----------
    original_value: float | np.ndarray
        The original value of the data.
    first_contract: bool
        The condition whether the row is existed in first or second contract in Transition scheme.
    prior_rows: np.ndarray
        The array that will be placed before the given value.
    post_rows:
        The array that will be placed after the given value.
    Returns
    -------
    adjusted_values: np.ndarray
        The array containin the adjusted value.

    """
    if isinstance(original_value, float):
        adjusted_values = original_value

    elif isinstance(original_value, np.ndarray) and first_contract:
        adjusted_values = np.concatenate((original_value, post_rows))

    elif isinstance(original_value, np.ndarray) and not first_contract:
        adjusted_values = np.concatenate((prior_rows, original_value))

    else:
        adjusted_values = original_value

    return adjusted_values


def adjusting_contract_arguments(arguments_dict: dict,
                                 project_years: np.ndarray,
                                 first_contract: bool,
                                 prior_rows: np.ndarray,
                                 post_rows: np.ndarray,
                                 ):
    """
    The function used to adjust the contract arguments before parsed to the transition contract.
    Parameters
    ----------
    arguments_dict: dict
        The dictionary containing the arguments of a contract.
    project_years: np.ndarray
        The array of the project years.
    first_contract: bool
        The condition whether the row is existed in first or second contract in Transition scheme.
    prior_rows: np.ndarray
        The array that will be placed before the given value.
    post_rows:
        The array that will be placed after the given value.
    Returns
    -------
    arguments_dict: dict
        The dictionary containing adjusted arguments.
    """
    if 'tax_rate' in arguments_dict.keys():
        arguments_dict['tax_rate'] = adjust_rows(original_value=arguments_dict['tax_rate'],
                                                 first_contract=first_contract,
                                                 prior_rows=prior_rows,
                                                 post_rows=post_rows, )

    if 'vat_rate' in arguments_dict.keys():
        arguments_dict['vat_rate'] = adjust_rows(original_value=arguments_dict['vat_rate'],
                                                 first_contract=first_contract,
                                                 prior_rows=prior_rows,
                                                 post_rows=post_rows, )

    if 'lbt_rate' in arguments_dict.keys():
        arguments_dict['lbt_rate'] = adjust_rows(original_value=arguments_dict['lbt_rate'],
                                                 first_contract=first_contract,
                                                 prior_rows=prior_rows,
                                                 post_rows=post_rows, )

    if 'inflation_rate' in arguments_dict.keys():
        arguments_dict['inflation_rate'] = adjust_rows(original_value=arguments_dict['inflation_rate'],
                                                       first_contract=first_contract,
                                                       prior_rows=prior_rows,
                                                       post_rows=post_rows, )

    if 'cum_production_split_offset' in arguments_dict.keys():
        arguments_dict['cum_production_split_offset'] = adjust_rows(
            original_value=arguments_dict['cum_production_split_offset'],
            first_contract=first_contract,
            prior_rows=prior_rows,
            post_rows=post_rows, )

    return arguments_dict


@dataclass
class Transition:
    """
    Dataclass that represents Transition contract.

    Parameters
    ----------
    contract1: CostRecovery | GrossSplit
        The object of the first contract.
    contract2: CostRecovery | GrossSplit
        The object of the second contract.
    argument_contract1: dict
        The argument of the first contract.
    argument_contract2: dict
        The argument of the second contract.
    """
    contract1: CostRecovery | GrossSplit
    contract2: CostRecovery | GrossSplit
    argument_contract1: dict
    argument_contract2: dict

    # Attributes to define later
    _contract1_transitioned: CostRecovery | GrossSplit = field(default=None, init=False, repr=False)
    _contract2_transitioned: CostRecovery | GrossSplit = field(default=None, init=False, repr=False)

    project_years: np.ndarray = field(default=None, init=False, repr=False)

    _oil_lifting: Lifting = field(default=None, init=False, repr=False)
    _gas_lifting: Lifting = field(default=None, init=False, repr=False)
    _sulfur_lifting: Lifting = field(default=None, init=False, repr=False)
    _electricity_lifting: Lifting = field(default=None, init=False, repr=False)
    _co2_lifting: Lifting = field(default=None, init=False, repr=False)

    _oil_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _gas_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _sulfur_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _electricity_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _co2_revenue: np.ndarray = field(default=None, init=False, repr=False)

    _oil_capital_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_capital_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_intangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_intangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_opex_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_opex_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_asr_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_asr_expenditures: np.ndarray = field(default=None, init=False, repr=False)

    _oil_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _gas_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _sulfur_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _electricity_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _co2_wap_price: np.ndarray = field(default=None, init=False, repr=False)

    _oil_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_unrec_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_unrec_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_ets: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_ets: np.ndarray = field(default=None, init=False, repr=False)
    _oil_gov_ets: np.ndarray = field(default=None, init=False, repr=False)
    _gas_gov_ets: np.ndarray = field(default=None, init=False, repr=False)
    _net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _ctr_net_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _ctr_ftp: np.ndarray = field(default=None, init=False, repr=False)
    _gov_ftp: np.ndarray = field(default=None, init=False, repr=False)
    _ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _government_take: np.ndarray = field(default=None, init=False, repr=False)

    # Consolidated Attributes
    _consolidated_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_government_take: np.ndarray = field(default=None, init=False, repr=False)

    @staticmethod
    def _parse_dataclass(
            contract: CostRecovery | GrossSplit,
            start_date_trans: date,
            end_date_trans: date,
            lifting: tuple,
            capital: tuple,
            intangible: tuple,
            opex: tuple,
            asr: tuple,
    ):

        # Condition where only one fluid is produced
        produced_fluid = [i.fluid_type for i in contract.lifting]

        if FluidType.OIL not in produced_fluid:
            oil_onstream_date = None
        else:
            oil_onstream_date = contract.oil_onstream_date
        if FluidType.GAS not in produced_fluid:
            gas_onstream_date = None
        else:
            gas_onstream_date = contract.gas_onstream_date

        if isinstance(contract, CostRecovery):
            new_contract = CostRecovery(
                start_date=start_date_trans,
                end_date=end_date_trans,
                oil_onstream_date=oil_onstream_date,
                gas_onstream_date=gas_onstream_date,
                lifting=lifting,
                capital_cost=capital,
                intangible_cost=intangible,
                opex=opex,
                asr_cost=asr,
                oil_ftp_is_available=contract.oil_ftp_is_available,
                oil_ftp_is_shared=contract.oil_ftp_is_shared,
                oil_ftp_portion=contract.oil_ftp_portion,
                gas_ftp_is_available=contract.gas_ftp_is_available,
                gas_ftp_is_shared=contract.gas_ftp_is_shared,
                gas_ftp_portion=contract.gas_ftp_portion,
                tax_split_type=contract.tax_split_type,
                condition_dict=contract.condition_dict,
                indicator_rc_icp_sliding=contract.indicator_rc_icp_sliding,
                oil_ctr_pretax_share=contract.oil_ctr_pretax_share,
                gas_ctr_pretax_share=contract.gas_ctr_pretax_share,
                oil_ic_rate=contract.oil_ic_rate,
                gas_ic_rate=contract.gas_ic_rate,
                ic_is_available=contract.ic_is_available,
                oil_cr_cap_rate=contract.oil_cr_cap_rate,
                gas_cr_cap_rate=contract.gas_cr_cap_rate,
                oil_dmo_volume_portion=contract.oil_dmo_volume_portion,
                oil_dmo_fee_portion=contract.oil_dmo_fee_portion,
                oil_dmo_holiday_duration=contract.oil_dmo_holiday_duration,
                gas_dmo_volume_portion=contract.gas_dmo_volume_portion,
                gas_dmo_fee_portion=contract.gas_dmo_fee_portion,
                gas_dmo_holiday_duration=contract.gas_dmo_holiday_duration)
        else:
            new_contract = GrossSplit(
                start_date=start_date_trans,
                end_date=end_date_trans,
                oil_onstream_date=oil_onstream_date,
                gas_onstream_date=gas_onstream_date,
                lifting=lifting,
                capital_cost=capital,
                intangible_cost=intangible,
                opex=opex,
                asr_cost=asr,
                field_status=contract.field_status,
                field_loc=contract.field_loc,
                res_depth=contract.res_depth,
                infra_avail=contract.infra_avail,
                res_type=contract.res_type,
                api_oil=contract.api_oil,
                domestic_use=contract.domestic_use,
                prod_stage=contract.prod_stage,
                co2_content=contract.co2_content,
                h2s_content=contract.h2s_content,
                base_split_ctr_oil=contract.base_split_ctr_oil,
                base_split_ctr_gas=contract.base_split_ctr_gas,
                split_ministry_disc=contract.split_ministry_disc,
                oil_dmo_volume_portion=contract.oil_dmo_volume_portion,
                oil_dmo_fee_portion=contract.oil_dmo_fee_portion,
                oil_dmo_holiday_duration=contract.oil_dmo_holiday_duration,
                gas_dmo_volume_portion=contract.gas_dmo_volume_portion,
                gas_dmo_fee_portion=contract.gas_dmo_fee_portion,
                gas_dmo_holiday_duration=contract.gas_dmo_holiday_duration,
            )

        return new_contract

    def run(self, unrec_portion: float = 0.0):
        # Defining the transition start date and end date
        start_date_trans = min([self.contract1.start_date, self.contract2.start_date])
        end_date_trans = max([self.contract1.end_date, self.contract2.end_date])

        # Condition where the first contract end on the 31 December and new contract start on  1 January
        condition = np.logical_and(self.contract1.end_date.day == 31, self.contract1.end_date.month == 12)
        if condition:
            # additional_year = 1
            additional_year = 0
            end_date_condition = 1
        else:
            additional_year = 0
            end_date_condition = 0

        project_years_trans = np.arange(start_date_trans.year, end_date_trans.year + 1 + additional_year)
        project_duration_trans = end_date_trans.year - start_date_trans.year + 1 + additional_year

        # Defining the array of years between the prior contract to the new contract
        years_to_new = np.arange(self.contract1.end_date.year + 1,
                                 self.contract2.end_date.year + 1 + additional_year - end_date_condition)
        years_to_prior = np.arange(self.contract1.start_date.year,
                                   self.contract2.start_date.year + additional_year - end_date_condition)

        len_to_new = len(years_to_new)
        len_to_prior = len(years_to_prior)

        # Defining the row gap and description between the prior contract and the new contract
        zeros_to_new = np.zeros_like(years_to_new, dtype=float)
        zeros_to_prior = np.zeros_like(years_to_prior, dtype=float)
        desc_to_new = ['-'] * len_to_new
        desc_to_prior = ['-'] * len_to_prior
        fluid_to_new = [FluidType.OIL] * len_to_new
        fluid_to_prior = [FluidType.OIL] * len_to_prior
        ic_to_new = [False] * len_to_new
        ic_to_prior = [False] * len_to_prior

        # Modifying the contract1
        # Changing the start_date and end_date of contract1
        new_lifting_1st = [
            Lifting(start_year=start_date_trans.year,
                    end_year=end_date_trans.year,
                    lifting_rate=np.concatenate((lift.lifting_rate, zeros_to_new)),
                    price=np.concatenate((lift.price, zeros_to_new)),
                    prod_year=np.concatenate((lift.prod_year, years_to_new)),
                    fluid_type=lift.fluid_type,
                    ghv=np.concatenate((lift.ghv, zeros_to_new)),
                    prod_rate=np.concatenate((lift.prod_rate, zeros_to_new)),
                    prod_rate_baseline=np.concatenate((lift.prod_rate_baseline, zeros_to_new)))
            for lift in self.contract1.lifting
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract1 tangible
        new_tang_1st = [
            CapitalCost(start_year=start_date_trans.year,
                        end_year=end_date_trans.year,
                        cost=np.concatenate((tang.cost, zeros_to_new)),
                        expense_year=np.concatenate((tang.expense_year, years_to_new)).astype(int),
                        cost_allocation=tang.cost_allocation + fluid_to_new,
                        description=tang.description + desc_to_new,
                        vat_portion=np.concatenate((tang.vat_portion, zeros_to_new)),
                        vat_discount=np.concatenate((tang.vat_discount, zeros_to_new)),
                        lbt_portion=np.concatenate((tang.lbt_portion, zeros_to_new)),
                        lbt_discount=np.concatenate((tang.lbt_discount, zeros_to_new)),
                        pis_year=np.concatenate((tang.pis_year, zeros_to_new)).astype(int),
                        salvage_value=np.concatenate((tang.salvage_value, zeros_to_new)),
                        useful_life=np.concatenate((tang.useful_life, zeros_to_new)),
                        depreciation_factor=np.concatenate((tang.depreciation_factor, zeros_to_new)),
                        is_ic_applied=tang.is_ic_applied + ic_to_new,)
            for tang in self.contract1.capital_cost
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract1 intangible
        new_intang_1st = [
            Intangible(start_year=start_date_trans.year,
                       end_year=end_date_trans.year,
                       cost=np.concatenate((intang.cost, zeros_to_new)),
                       expense_year=np.concatenate((intang.expense_year, years_to_new)).astype(int),
                       cost_allocation=intang.cost_allocation + fluid_to_new,
                       description=intang.description + desc_to_new,
                       vat_portion=np.concatenate((intang.vat_portion, zeros_to_new)),
                       vat_discount=np.concatenate((intang.vat_discount, zeros_to_new)),
                       lbt_portion=np.concatenate((intang.lbt_portion, zeros_to_new)),
                       lbt_discount=np.concatenate((intang.lbt_discount, zeros_to_new)))
            for intang in self.contract1.intangible_cost
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract1 opex
        new_opex_1st = [
            OPEX(start_year=start_date_trans.year,
                 end_year=end_date_trans.year,
                 expense_year=np.concatenate((opx.expense_year, years_to_new)).astype(int),
                 cost_allocation=opx.cost_allocation + fluid_to_new,
                 description=opx.description + desc_to_new,
                 fixed_cost=np.concatenate((opx.fixed_cost, zeros_to_new)),
                 prod_rate=np.concatenate((opx.prod_rate, zeros_to_new)),
                 cost_per_volume=np.concatenate((opx.cost_per_volume, zeros_to_new)),
                 vat_portion=np.concatenate((opx.vat_portion, zeros_to_new)),
                 vat_discount=np.concatenate((opx.vat_discount, zeros_to_new)),
                 lbt_portion=np.concatenate((opx.lbt_portion, zeros_to_new)),
                 lbt_discount=np.concatenate((opx.lbt_discount, zeros_to_new)),)
            for opx in self.contract1.opex
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract1 asr
        new_asr_1st = [
            ASR(start_year=start_date_trans.year,
                end_year=end_date_trans.year,
                cost=np.concatenate((asr.cost, zeros_to_new)),
                expense_year=np.concatenate((asr.expense_year, years_to_new)).astype(int),
                cost_allocation=asr.cost_allocation + fluid_to_new,
                description=asr.description + desc_to_new,
                vat_portion=np.concatenate((asr.vat_portion, zeros_to_new)),
                vat_discount=np.concatenate((asr.vat_discount, zeros_to_new)),
                lbt_portion=np.concatenate((asr.lbt_portion, zeros_to_new)),
                lbt_discount=np.concatenate((asr.lbt_discount, zeros_to_new)))
            for asr in self.contract1.asr_cost
        ]

        # Modifying the contract2
        # Changing the start_date and end_date of contract2

        # Concatenating the zeros_to_new and years_to_new to the contract2 lifting
        new_lifting_2nd = [
            Lifting(start_year=start_date_trans.year,
                    end_year=end_date_trans.year,
                    lifting_rate=np.concatenate((zeros_to_prior, lift.lifting_rate)),
                    price=np.concatenate((zeros_to_prior, lift.price)),
                    prod_year=np.concatenate((years_to_prior, lift.prod_year)).astype(int),
                    fluid_type=lift.fluid_type,
                    ghv=np.concatenate((zeros_to_prior, lift.ghv)),
                    prod_rate=np.concatenate((zeros_to_prior, lift.prod_rate)),
                    prod_rate_baseline=np.concatenate((zeros_to_prior, lift.prod_rate_baseline)))
            for lift in self.contract2.lifting
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract2 tangible
        new_tang_2nd = [
            CapitalCost(start_year=start_date_trans.year,
                        end_year=end_date_trans.year,
                        cost=np.concatenate((zeros_to_prior, tang.cost)),
                        expense_year=np.concatenate((years_to_prior, tang.expense_year)).astype(int),
                        cost_allocation=fluid_to_prior + tang.cost_allocation,
                        description=desc_to_prior + tang.description,
                        vat_portion=np.concatenate((zeros_to_prior, tang.vat_portion)),
                        vat_discount=np.concatenate((zeros_to_prior, tang.vat_discount)),
                        lbt_portion=np.concatenate((zeros_to_prior, tang.lbt_portion)),
                        lbt_discount=np.concatenate((zeros_to_prior, tang.lbt_discount)),
                        pis_year=np.concatenate((zeros_to_prior, tang.pis_year)).astype(int),
                        salvage_value=np.concatenate((zeros_to_prior, tang.salvage_value)),
                        useful_life=np.concatenate((zeros_to_prior, tang.useful_life)),
                        depreciation_factor=np.concatenate((zeros_to_prior, tang.depreciation_factor)),
                        is_ic_applied=ic_to_prior + tang.is_ic_applied)
            for tang in self.contract2.capital_cost
        ]
        # Concatenating the zeros_to_new and years_to_new to the contract2 intangible
        new_intang_2nd = [
            Intangible(start_year=start_date_trans.year,
                       end_year=end_date_trans.year,
                       cost=np.concatenate((zeros_to_prior, intang.cost)),
                       expense_year=np.concatenate((years_to_prior, intang.expense_year)).astype(int),
                       cost_allocation=fluid_to_prior + intang.cost_allocation,
                       description=desc_to_prior + intang.description,
                       vat_portion=np.concatenate((zeros_to_prior, intang.vat_portion)),
                       vat_discount=np.concatenate((zeros_to_prior, intang.vat_discount)),
                       lbt_portion=np.concatenate((zeros_to_prior, intang.lbt_portion)),
                       lbt_discount=np.concatenate((zeros_to_prior, intang.lbt_discount)))
            for intang in self.contract2.intangible_cost
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract2 opex
        new_opex_2nd = [
            OPEX(start_year=start_date_trans.year,
                 end_year=end_date_trans.year,
                 expense_year=np.concatenate((years_to_prior, opx.expense_year)).astype(int),
                 cost_allocation=fluid_to_prior + opx.cost_allocation,
                 description=desc_to_prior + opx.description,
                 fixed_cost=np.concatenate((zeros_to_prior, opx.fixed_cost)),
                 prod_rate=np.concatenate((zeros_to_prior, opx.prod_rate)),
                 cost_per_volume=np.concatenate((zeros_to_prior, opx.cost_per_volume)),
                 vat_portion=np.concatenate((zeros_to_prior, opx.vat_portion)),
                 vat_discount=np.concatenate((zeros_to_prior, opx.vat_discount)),
                 lbt_portion=np.concatenate((zeros_to_prior, opx.lbt_portion)),
                 lbt_discount=np.concatenate((zeros_to_prior, opx.lbt_discount)))
            for opx in self.contract2.opex
        ]

        # Concatenating the zeros_to_new and years_to_new to the contract2 asr
        new_asr_2nd = [
            ASR(start_year=start_date_trans.year,
                end_year=end_date_trans.year,
                cost=np.concatenate((zeros_to_prior, asr.cost)),
                expense_year=np.concatenate((years_to_prior, asr.expense_year)).astype(int),
                cost_allocation=fluid_to_prior + asr.cost_allocation,
                description=desc_to_prior + asr.description,
                vat_portion=np.concatenate((zeros_to_prior, asr.vat_portion)),
                vat_discount=np.concatenate((zeros_to_prior, asr.vat_discount)),
                lbt_portion=np.concatenate((zeros_to_prior, asr.lbt_portion)),
                lbt_discount=np.concatenate((zeros_to_prior, asr.lbt_discount)))
            for asr in self.contract2.asr_cost
        ]

        # Parsing the attributes to the new object of contract
        contract1_new = self._parse_dataclass(
            contract=self.contract1,
            start_date_trans=start_date_trans,
            end_date_trans=end_date_trans,
            lifting=tuple(new_lifting_1st),
            capital=tuple(new_tang_1st),
            intangible=tuple(new_intang_1st),
            opex=tuple(new_opex_1st),
            asr=tuple(new_asr_1st),
        )

        contract2_new = self._parse_dataclass(
            contract=self.contract2,
            start_date_trans=start_date_trans,
            end_date_trans=end_date_trans,
            lifting=tuple(new_lifting_2nd),
            capital=tuple(new_tang_2nd),
            intangible=tuple(new_intang_2nd),
            opex=tuple(new_opex_2nd),
            asr=tuple(new_asr_2nd),
        )

        # Adjusting the contract arguments
        new_argument_contract1 = adjusting_contract_arguments(arguments_dict=self.argument_contract1,
                                                              project_years=project_years_trans,
                                                              first_contract=True,
                                                              prior_rows=zeros_to_prior,
                                                              post_rows=zeros_to_new,)

        new_argument_contract2 = adjusting_contract_arguments(arguments_dict=self.argument_contract2,
                                                              project_years=project_years_trans,
                                                              first_contract=False,
                                                              prior_rows=zeros_to_prior,
                                                              post_rows=zeros_to_new,)

        # Executing the new contract
        # for i in self.argument_contract1.keys():
        #     print(i, ': ', self.argument_contract1[i])
        # input()

        contract1_new.run(**new_argument_contract1)
        contract2_new.run(**new_argument_contract2)

        # Defining the unrecoverable cost from the prior contract
        # Condition if the contract is Cost Recovery
        if isinstance(contract1_new, CostRecovery):
            latest_unrec = contract1_new._consolidated_unrecovered_after_transfer[-1]
        # Condition if the contract is Gross Split
        else:
            latest_unrec = contract1_new._consolidated_carward_cost_aftertf[-1]
        unrec_cost_prior = np.zeros_like(contract2_new.project_years, dtype=float)
        unrec_cost_prior[len(years_to_prior)] = latest_unrec

        # Calculating the Unrecoverable cost in the transitioned cashflow
        unrec_trans = psc_tools.get_unrecovered_cost(revenue=contract2_new._consolidated_taxable_income,
                                                     depreciation=unrec_cost_prior,
                                                     non_capital=np.zeros_like(contract2_new.project_years,
                                                                               dtype=float),
                                                     ftp_ctr=np.zeros_like(contract2_new.project_years, dtype=float),
                                                     ftp_gov=np.zeros_like(contract2_new.project_years, dtype=float),
                                                     ic=np.zeros_like(contract2_new.project_years, dtype=float))

        cost_to_be_recovered_trans = psc_tools.get_cost_to_be_recovered(unrecovered_cost=unrec_trans)

        cost_recovery_trans = CostRecovery._get_cost_recovery(
            revenue=contract2_new._consolidated_taxable_income,
            ftp=np.zeros_like(contract2_new.project_years, dtype=float),
            ic=np.zeros_like(contract2_new.project_years, dtype=float),
            depreciation=unrec_cost_prior,
            non_capital=np.zeros_like(contract2_new.project_years, dtype=float),
            cost_to_be_recovered=cost_to_be_recovered_trans,
            cr_cap_rate=1.0)

        # Calculate taxable income after considering transferred unrecoverable cost from contract1
        taxable_income_trans = CostRecovery._get_ets_before_transfer(revenue=contract2_new._consolidated_taxable_income,
                                                                     ftp_ctr=np.zeros_like(contract2_new.project_years,
                                                                                           dtype=float),
                                                                     ftp_gov=np.zeros_like(contract2_new.project_years,
                                                                                           dtype=float),
                                                                     ic=np.zeros_like(contract2_new.project_years,
                                                                                      dtype=float),
                                                                     cost_recovery=cost_recovery_trans)

        # Calculate tax payment of adjusted taxable income
        tax_payment_transition = taxable_income_trans * contract2_new._tax_rate_arr

        # Calculate new ctr net share
        contract2_new._consolidated_ctr_net_share = (contract2_new._consolidated_taxable_income -
                                                     (unrec_cost_prior * unrec_portion) -
                                                     tax_payment_transition)

        # Calculate new cashflow of contract2
        contract2_new._consolidated_cashflow = (contract2_new._consolidated_cashflow +
                                                contract2_new._consolidated_tax_payment -
                                                tax_payment_transition
                                                )

        # Calculate new government take
        contract2_new._consolidated_government_take = (contract2_new._consolidated_government_take -
                                                       contract2_new._consolidated_tax_payment +
                                                       tax_payment_transition)

        contract2_new._consolidated_tax_payment = tax_payment_transition

        # Parsing new contract to the attributes of Transition dataclass
        self._contract1_transitioned = contract1_new
        self._contract2_transitioned = contract2_new

        # Populating transition attributes
        # Lifting
        self._oil_lifting = (self._contract1_transitioned._oil_lifting +
                             self._contract2_transitioned._oil_lifting)

        self._gas_lifting = (self._contract1_transitioned._gas_lifting +
                             self._contract2_transitioned._gas_lifting)

        self._sulfur_lifting = (self._contract1_transitioned._sulfur_lifting +
                                self._contract2_transitioned._sulfur_lifting)

        self._electricity_lifting = (self._contract1_transitioned._electricity_lifting +
                                     self._contract2_transitioned._electricity_lifting)

        self._co2_lifting = (self._contract1_transitioned._co2_lifting +
                             self._contract2_transitioned._co2_lifting)

        # Revenue
        self._oil_revenue = (self._contract1_transitioned._oil_revenue +
                             self._contract2_transitioned._oil_revenue)

        self._gas_revenue = (self._contract1_transitioned._gas_revenue +
                             self._contract2_transitioned._gas_revenue)

        self._sulfur_revenue = (self._contract1_transitioned._sulfur_revenue +
                                self._contract2_transitioned._sulfur_revenue)

        self._electricity_revenue = (self._contract1_transitioned._electricity_revenue +
                                     self._contract2_transitioned._electricity_revenue)

        self._co2_revenue = (self._contract1_transitioned._co2_revenue +
                             self._contract2_transitioned._co2_revenue)

        # Tangible
        self._oil_capital_expenditures = (self._contract1_transitioned._oil_capital_expenditures +
                                          self._contract2_transitioned._oil_capital_expenditures)

        self._gas_capital_expenditures = (self._contract1_transitioned._gas_capital_expenditures +
                                          self._contract2_transitioned._gas_capital_expenditures)

        # Intangible
        self._oil_intangible_expenditures = (self._contract1_transitioned._oil_intangible_expenditures +
                                             self._contract2_transitioned._oil_intangible_expenditures)

        self._gas_intangible_expenditures = (self._contract1_transitioned._gas_intangible_expenditures +
                                             self._contract2_transitioned._gas_intangible_expenditures)

        # Opex
        self._oil_opex_expenditures = (self._contract1_transitioned._oil_opex_expenditures +
                                       self._contract2_transitioned._oil_opex_expenditures)

        self._gas_opex_expenditures = (self._contract1_transitioned._gas_opex_expenditures +
                                       self._contract2_transitioned._gas_opex_expenditures)

        # ASR
        self._oil_asr_expenditures = (self._contract1_transitioned._oil_asr_expenditures +
                                      self._contract2_transitioned._oil_asr_expenditures)

        self._gas_asr_expenditures = (self._contract1_transitioned._gas_asr_expenditures +
                                      self._contract2_transitioned._gas_asr_expenditures)

        # WAP Price
        self._oil_wap_price = self._contract2_transitioned._oil_wap_price
        self._gas_wap_price = self._contract2_transitioned._gas_wap_price
        self._sulfur_wap_price = self._contract2_transitioned._sulfur_wap_price
        self._electricity_wap_price = self._contract2_transitioned._electricity_wap_price
        self._co2_wap_price = self._contract2_transitioned._co2_wap_price

        # Sunk Cost
        self._oil_sunk_cost = self._contract1_transitioned._oil_sunk_cost + self._contract2_transitioned._oil_sunk_cost
        self._gas_sunk_cost = self._contract1_transitioned._gas_sunk_cost + self._contract2_transitioned._gas_sunk_cost

        # Deductible Cost / Cost Recovery
        _oil_deductible_cost_1 = None
        _gas_deductible_cost_1 = None
        _oil_deductible_cost_2 = None
        _gas_deductible_cost_2 = None

        _oil_unrec_cost_1 = None
        _gas_unrec_cost_1 = None
        _oil_unrec_cost_2 = None
        _gas_unrec_cost_2 = None

        _oil_ctr_ets_1 = None
        _gas_ctr_ets_1 = None
        _oil_ctr_ets_2 = None
        _gas_ctr_ets_2 = None

        _oil_gov_ets_1 = None
        _gas_gov_ets_1 = None
        _oil_gov_ets_2 = None
        _gas_gov_ets_2 = None

        _ctr_ftp_1 = None
        _gov_ftp_1 = None
        _ctr_ftp_2 = None
        _gov_ftp_2 = None

        # Determining variables if the type of Contract1 is Cost Recovery
        if isinstance(self.contract1, CostRecovery):
            _oil_deductible_cost_1 = self._contract1_transitioned._oil_cost_recovery_after_tf
            _gas_deductible_cost_1 = self._contract1_transitioned._gas_cost_recovery_after_tf
            _oil_unrec_cost_1 = self._contract1_transitioned._oil_unrecovered_after_transfer
            _gas_unrec_cost_1 = self._contract1_transitioned._gas_unrecovered_after_transfer
            _oil_ctr_ets_1 = self._contract1_transitioned._oil_ets_after_transfer
            _gas_ctr_ets_1 = self._contract1_transitioned._gas_ets_after_transfer
            _oil_gov_ets_1 = self._contract1_transitioned._oil_government_share
            _gas_gov_ets_1 = self._contract1_transitioned._gas_government_share
            _ctr_ftp_1 = self._contract1_transitioned._consolidated_ftp_ctr
            _gov_ftp_1 = self._contract1_transitioned._consolidated_ftp_gov

        # Determining variables if the type of Contract1 is Gross Split
        elif isinstance(self.contract1, GrossSplit):
            _oil_deductible_cost_1 = self._contract1_transitioned._oil_deductible_cost
            _gas_deductible_cost_1 = self._contract1_transitioned._gas_deductible_cost
            _oil_unrec_cost_1 = self._contract1_transitioned._oil_carward_deduct_cost
            _gas_unrec_cost_1 = self._contract1_transitioned._gas_carward_deduct_cost
            _oil_ctr_ets_1 = self._contract1_transitioned._oil_ctr_share_after_transfer
            _gas_ctr_ets_1 = self._contract1_transitioned._gas_ctr_share_after_transfer
            _oil_gov_ets_1 = np.zeros_like(self._contract1_transitioned.project_years)
            _gas_gov_ets_1 = np.zeros_like(self._contract1_transitioned.project_years)
            _ctr_ftp_1 = np.zeros_like(self._contract1_transitioned.project_years, dtype=float)
            _gov_ftp_1 = self._contract1_transitioned._consolidated_gov_share_before_tf

        # Determining variables if the type of Contract2 is Cost Recovery
        if isinstance(self.contract2, CostRecovery):
            _oil_deductible_cost_2 = self._contract2_transitioned._oil_cost_recovery_after_tf
            _gas_deductible_cost_2 = self._contract2_transitioned._gas_cost_recovery_after_tf
            _oil_unrec_cost_2 = self._contract2_transitioned._oil_unrecovered_after_transfer
            _gas_unrec_cost_2 = self._contract2_transitioned._gas_unrecovered_after_transfer
            _oil_ctr_ets_2 = self._contract2_transitioned._oil_ets_after_transfer
            _gas_ctr_ets_2 = self._contract2_transitioned._gas_ets_after_transfer
            _oil_gov_ets_2 = self._contract2_transitioned._oil_government_share
            _gas_gov_ets_2 = self._contract2_transitioned._gas_government_share
            _ctr_ftp_2 = self._contract2_transitioned._consolidated_ftp_ctr
            _gov_ftp_2 = self._contract2_transitioned._consolidated_ftp_gov

        # Determining variables if the type of Contract2 is Gross Split
        elif isinstance(self.contract2, GrossSplit):
            _oil_deductible_cost_2 = self._contract2_transitioned._oil_deductible_cost
            _gas_deductible_cost_2 = self._contract2_transitioned._gas_deductible_cost
            _oil_unrec_cost_2 = self._contract2_transitioned._oil_carward_deduct_cost
            _gas_unrec_cost_2 = self._contract2_transitioned._gas_carward_deduct_cost
            _oil_ctr_ets_2 = self._contract2_transitioned._oil_ctr_share_after_transfer
            _gas_ctr_ets_2 = self._contract2_transitioned._gas_ctr_share_after_transfer
            _oil_gov_ets_2 = np.zeros_like(self._contract2_transitioned.project_years)
            _gas_gov_ets_2 = np.zeros_like(self._contract2_transitioned.project_years)
            _ctr_ftp_2 = np.zeros_like(self._contract2_transitioned.project_years, dtype=float)
            _gov_ftp_2 = self._contract2_transitioned._consolidated_gov_share_before_tf

        # Deductible Cost / Cost Recovery
        self._oil_deductible_cost = _oil_deductible_cost_1 + _oil_deductible_cost_2
        self._gas_deductible_cost = _gas_deductible_cost_1 + _gas_deductible_cost_2

        # Carry Forward Cost / Unrecoverable Cost
        self._oil_unrec_cost = _oil_unrec_cost_2
        self._gas_unrec_cost = _gas_unrec_cost_2

        # ETS
        self._oil_ctr_ets = _oil_ctr_ets_1 + _oil_ctr_ets_2
        self._gas_ctr_ets = _gas_ctr_ets_1 + _gas_ctr_ets_2

        self._oil_gov_ets = _oil_gov_ets_1 + _oil_gov_ets_2
        self._gas_gov_ets = _gas_gov_ets_1 + _gas_gov_ets_2

        # Contractor Net Operating Profit
        self._net_operating_profit = (self._contract1_transitioned._consolidated_ctr_net_share +
                                      self._contract2_transitioned._consolidated_ctr_net_share)

        # Contractor Net Cash Flow
        self._ctr_net_cashflow = (self._contract1_transitioned._consolidated_cashflow +
                                  self._contract2_transitioned._consolidated_cashflow)

        # FTP
        self._ctr_ftp = _ctr_ftp_1 + _ctr_ftp_2
        self._gov_ftp = _gov_ftp_1 + _gov_ftp_2

        # DDMO
        self._ddmo = (self._contract1_transitioned._consolidated_ddmo +
                      self._contract2_transitioned._consolidated_ddmo)

        # Tax
        self._tax_payment = (self._contract1_transitioned._consolidated_tax_payment +
                             self._contract2_transitioned._consolidated_tax_payment)

        # Government Take
        self._government_take = (self._contract1_transitioned._consolidated_government_take +
                                 self._contract2_transitioned._consolidated_government_take)

        # Consolidated Attributes
        # Consolidated Revenue
        self._consolidated_revenue = self._oil_revenue + self._gas_revenue

        # Consolidated Sunk Cost
        self._consolidated_sunk_cost = self._oil_sunk_cost + self._gas_sunk_cost

        # Consolidated Cash Flow
        self._consolidated_cashflow = (self._contract1_transitioned._consolidated_cashflow +
                                       self._contract2_transitioned._consolidated_cashflow)

        # Consolidated DDMO
        self._consolidated_ddmo = (self._contract1_transitioned._consolidated_ddmo +
                                   self._contract2_transitioned._consolidated_ddmo)

        # Consolidated Tax Payment
        self._consolidated_tax_payment = (self._contract1_transitioned._consolidated_tax_payment +
                                          self._contract2_transitioned._consolidated_tax_payment)

        # Consolidated Government Take
        self._consolidated_government_take = (self._contract1_transitioned._consolidated_government_take +
                                              self._contract2_transitioned._consolidated_government_take)

        # Project Years
        self.project_years = np.copy(self._contract1_transitioned.project_years)



"""
Prepares lifting data and calculate the revenue.
"""

import numpy as np
from dataclasses import dataclass, field

from pyscnomics.econ.selection import FluidType
from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR


class LiftingException(Exception):
    """Exception to be raised if class Lifting is misused"""

    pass


@dataclass
class Lifting:
    """
    Create an object that represents lifting.

    Parameters
    ----------
    start_year: int
        The start year of the project.
    end_year: int
        The end year of the project.
    lifting_rate: np.ndarray
        The lifting data of a particular fluid type.
    price: np.ndarray
        The associated price of a particular fluid type.
    prod_year: np.ndarray
        The production year of a particular fluid type.
    fluid_type: FluidType
        The fluid type.
    ghv: np.ndarray
        The value of ghv of a particular fluid type (default value = 1).
    prod_rate: np.ndarray
        The production rate of a particular fluid type.

    Notes
    -----
    (1) The unit used in the lifting rate should be in M unit, where the M is stands for 1000.
        While the price should be in US Dollar (USD).

    (2) The corresponding lifting unit for each fluid type are as the following:
        - Oil: MBBL
        - Gas: BSCF
        - LPG Propane: M-Ton (MT)
        - LPG Butane: M-Ton (MT)
        - Sulfur: M-Ton (MT)
        - CO2: M-Ton (MT)
        - Electricity: M-Unit

    (3) The corresponding price unit for each fluid type are as the following:
        - Oil: USD/BBL
        - Gas: USD/MM-British Thermal Unit (MMBTU), while the GHV will be: M-BTU/Standard Cubic Feet (SCF)
        - LPG Propane: USD/MT
        - LPG Butane: USD/MT
        - Sulfur: USD/MT
        - CO2: USD/MT
        - Electricity: USD/MT
    """

    start_year: int
    end_year: int
    lifting_rate: np.ndarray
    price: np.ndarray
    prod_year: np.ndarray
    fluid_type: FluidType = field(default=FluidType.OIL)
    ghv: np.ndarray = field(default=None)
    prod_rate: np.ndarray = field(default=None)
    prod_rate_baseline: np.ndarray = field(default=None)

    # Attribute to be defined later on
    project_duration: int = field(default=None, init=False)
    project_years: np.ndarray = field(default=None, init=False, repr=False)
    prod_rate_total: np.ndarray = field(default=None, init=False)

    def __post_init__(self):
        # Prepare attribute lifting rate
        if not isinstance(self.lifting_rate, np.ndarray):
            raise LiftingException(
                f"Attribute lifting rate must be provided as numpy ndarray, "
                f"not as {self.lifting_rate.__class__.__qualname__}."
            )

        self.lifting_rate = self.lifting_rate.astype(np.float64)

        # Prepare attribute price
        if not isinstance(self.price, np.ndarray):
            raise LiftingException(
                f"Attribute price must be provided as numpy ndarray, "
                f"not as {self.price.__class__.__qualname__}."
            )

        self.price = self.price.astype(np.float64)

        # Prepare attribute prod_year
        if not isinstance(self.prod_year, np.ndarray):
            raise LiftingException(
                f"Attribute prod_year must be provided as numpy ndarray, "
                f"not as {self.prod_year.__class__.__qualname__}."
            )

        self.prod_year = self.prod_year.astype(np.int64)

        # Prepare attribute ghv
        if self.ghv is None:
            self.ghv = np.ones_like(self.lifting_rate, dtype=np.float64)
        else:
            if not isinstance(self.ghv, np.ndarray):
                raise LiftingException(
                    f"Attribute ghv must be provided as numpy ndarray, "
                    f"not as {self.ghv.__class__.__qualname__}."
                )

        self.ghv = self.ghv.astype(np.float64)

        # Prepare attribute prod rate
        if self.prod_rate is None:
            self.prod_rate = self.lifting_rate.copy()
        else:
            if not isinstance(self.prod_rate, np.ndarray):
                raise LiftingException(
                    f"Attribute prod rate must be provided as numpy ndarray, "
                    f"not as {self.prod_rate.__class__.__qualname__}."
                )

            prod_rate_indices = np.argwhere(self.prod_rate < self.lifting_rate).ravel()

            if len(prod_rate_indices) > 0:
                raise LiftingException(
                    f"Lifting rate is larger than the production rate at production year: "
                    f"{self.prod_year[prod_rate_indices]}"
                )

        self.prod_rate = self.prod_rate.astype(np.float64)

        # Prepare attribute prod_rate_baseline
        if self.prod_rate_baseline is None:
            self.prod_rate_baseline = np.zeros_like(self.lifting_rate, dtype=np.float64)
        else:
            if not isinstance(self.prod_rate_baseline, np.ndarray):
                raise LiftingException(
                    f"Attribute prod_rate_baseline must be provided as numpy ndarray, "
                    f"not as {self.prod_rate_baseline.__class__.__qualname__}."
                )

        self.prod_rate_baseline = self.prod_rate_baseline.astype(np.float64)

        # Check for unequal length of array data
        arr_length = len(self.lifting_rate)

        if not all(
            len(arr) == arr_length
            for arr in [self.price, self.ghv, self.prod_year, self.prod_rate, self.prod_rate_baseline]
        ):
            raise LiftingException(
                f"Unequal length of array: "
                f"lifting_rate: {len(self.lifting_rate)}, "
                f"price: {len(self.price)}, "
                f"prod_year: {len(self.prod_year)}, "
                f"ghv: {len(self.ghv)}, "
                f"prod_rate: {len(self.prod_rate)}, "
                f"prod_rate_baseline: {len(self.prod_rate_baseline)}."
            )

        # Prepare attribute prod_rate_total
        self.prod_rate_total = self.prod_rate + self.prod_rate_baseline

        # Prepare attribute project_years and project_duration
        # Raise an error message: start_year is after end_year
        if self.end_year >= self.start_year:
            self.project_duration = self.end_year - self.start_year + 1
            self.project_years = np.arange(self.start_year, self.end_year + 1, 1)

        else:
            raise LiftingException(
                f"start year {self.start_year} is after the end year: {self.end_year}"
            )

        # Raise an error message: prod_year is before the start year of the project
        if np.min(self.prod_year) < self.start_year:
            raise LiftingException(
                f"The production year ({np.min(self.prod_year)}) "
                f"is before the start year of the project ({self.start_year})"
            )

        # Raise an error message: prod_year is after the end year of the project
        if np.max(self.prod_year) > self.end_year:
            raise LiftingException(
                f"The production year ({np.max(self.prod_year)}) "
                f"is after the end year of the project ({self.end_year})"
            )

    def revenue(self) -> np.ndarray:
        """
        Calculate the revenue of a particular fluid type.

        Returns
        -------
        rev: np.ndarray
            The revenue of a particular fluid type.

        Notes
        -----
        The revenue is calculated as follows: revenue = lifting rate * price * ghv.
        The function np.bincount() is used to align the revenue elements with its
        correponding year.
        """

        rev = self.lifting_rate * self.price * self.ghv
        rev_update = np.bincount(self.prod_year - self.start_year, weights=rev)

        # Modify revenues, acoounting for project duration
        zeros = np.zeros(self.project_duration - len(rev_update))
        rev_update = np.concatenate((rev_update, zeros))

        return rev_update

    def _get_array(self, target_param: np.ndarray) -> np.ndarray:
        """
        Create an array of target_param.

        Parameters
        ----------
        target_param: np.ndarray
            An array containing the parameters to be weighted.

        Returns
        -------
        np.ndarray
            An array with values weighted by `target_param`, aligned with production years.

        Notes
        -----
        (1) Function np.bincount() is used to align the target_param according to
            its corresponding prod year,
        (2) If len(param_arr) < project_duration, then add the remaining elements
            with zeros.
        """
        param_arr = np.bincount(self.prod_year - self.start_year, weights=target_param)
        zeros = np.zeros(self.project_duration - len(param_arr))
        param_arr = np.concatenate((param_arr, zeros))

        return param_arr

    def get_lifting_rate_arr(self) -> np.ndarray:
        """
        Create an array of lifting rate according to the corresponding production year.

        Returns
        -------
        np.ndarray
            The array of lifting rate with length equals to project duration.

        Notes
        -----
        Array of lifting rate is generated by calling the private method self._get_array().
        """
        return self._get_array(target_param=self.lifting_rate * self.ghv)

    def get_prod_rate_arr(self) -> np.ndarray:
        """
        Create an array of production rate according to the corresponding production year.

        Returns
        -------
        np.ndarray
            The array of lifting rate with length equals to project duration.

        Notes
        -----
        Array of lifting rate is generated by calling the private method self._get_array().
        """
        return self._get_array(target_param=self.prod_rate)

    def get_prod_rate_total_arr(self) -> np.ndarray:
        """
        Create an array of total production rate (production rate + production rate baseline)
        aligned with the corresponding production year.

        Returns
        -------
        np.ndarray
            The array of total production rate (production rate + production rate baseline)
            with length equals to the project duration.

        Notes
        -----
        Array of total production rate is generated by calling the private method self._get_array().
        """
        return self._get_array(target_param=self.prod_rate_total)

    def get_price_arr(self) -> np.ndarray:
        """
        Create an array of price according to the corresponding production year.

        Returns
        -------
        np.ndarray
            The array of price with length equals to project duration.

        Notes
        -----
        Array of price is generated by calling the private method self._get_array().
        """
        return self._get_array(target_param=self.price)

    def get_lifting_ghv_arr(self) -> np.ndarray:
        """
        Create an array of ghv according to the corresponding production year.

        Returns
        -------
        np.ndarray
            The array of lifting rate with length equals to project duration.

        Notes
        -----
        Array of ghv is generated by calling the private method self._get_array().
        """
        return self._get_array(target_param=self.ghv)

    def __len__(self):
        return self.project_duration

    def __eq__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return all(
                (
                    self.fluid_type == other.fluid_type,
                    self.start_year == other.start_year,
                    self.end_year == other.end_year,
                    np.allclose(self.lifting_rate, other.lifting_rate),
                    np.allclose(self.price, other.price),
                    np.allclose(self.prod_year, other.prod_year),
                    np.allclose(self.ghv, other.ghv),
                    np.allclose(self.prod_rate, other.prod_rate),
                    np.allclose(self.prod_rate_baseline, other.prod_rate_baseline),
                )
            )

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            return np.allclose(np.sum(self.revenue()), other)

        else:
            return False

    def __lt__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return np.sum(self.revenue()) < np.sum(other.revenue())

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.revenue()) < other

        else:
            raise LiftingException(
                f"Must compare an instance of Lifting with another instance of Lifting, "
                f"an integer, or a float"
            )

    def __le__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return np.sum(self.revenue()) <= np.sum(other.revenue())

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.revenue()) <= other

        else:
            raise LiftingException(
                f"Must compare an instance of Lifting with another instance of Lifting, "
                f"an integer, or a float"
            )

    def __gt__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return np.sum(self.revenue()) > np.sum(other.revenue())

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.revenue()) > other

        else:
            raise LiftingException(
                f"Must compare an instance of Lifting with another instance of Lifting, "
                f"an integer, or a float"
            )

    def __ge__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return np.sum(self.revenue()) >= np.sum(other.revenue())

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.revenue()) >= other

        else:
            raise LiftingException(
                f"Must compare an instance of Lifting with another instance of Lifting, "
                f"an integer, or a float"
            )

    def __add__(self, other):
        # Between an instance of Lifting with another instance of Lifting
        if isinstance(other, Lifting):
            # Raise exception error if self.cost_allocation is not equal to other.cost_allocation
            if self.fluid_type != other.fluid_type:
                raise LiftingException(
                    "Cost allocation is not equal. "
                    f"First instance is {self.fluid_type}, "
                    f"second instance is {other.fluid_type} "
                )

            else:
                start_year = min(self.start_year, other.start_year)
                end_year = max(self.end_year, other.end_year)
                lifting_rate = np.concatenate((self.lifting_rate, other.lifting_rate))
                price = np.concatenate((self.price, other.price))
                prod_year = np.concatenate((self.prod_year, other.prod_year))
                ghv = np.concatenate((self.ghv, other.ghv))
                prod_rate = np.concatenate((self.prod_rate, other.prod_rate))
                prod_rate_baseline = np.concatenate((self.prod_rate_baseline, other.prod_rate_baseline))

                return Lifting(
                    start_year=start_year,
                    end_year=end_year,
                    lifting_rate=lifting_rate,
                    price=price,
                    prod_year=prod_year,
                    fluid_type=self.fluid_type,
                    ghv=ghv,
                    prod_rate=prod_rate,
                    prod_rate_baseline=prod_rate_baseline,
                )

        # Between an instance of Lifting with an integer/float
        elif isinstance(other, (int, float)):
            return self.revenue() + other

        else:
            raise LiftingException(
                f"Must add an instance of Lifting with another instance "
                f"of Lifting or int/float "
                f"{other}({other.__class__.__qualname__}) is not an instance of "
                f"Lifting or int/float"
            )

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        # If "other" is an instance of Lifting object
        if isinstance(other, Lifting):
            # Configure the minimum and maximum values of the start and end year, respectively
            start_year = min(self.start_year, other.start_year)
            end_year = max(self.end_year, other.end_year)

            self_revenue = self.revenue().copy()
            other_revenue = other.revenue().copy()

            # If the length of self_revenue data is less than project duration
            if len(self_revenue) < end_year - start_year + 1:
                # Modify the size of self_revenue data
                self_revenue.resize(end_year - start_year + 1, refcheck=False)

                # If self.start_year > other.start_year, roll the array n steps to the right;
                # n = self.start_year - start_year
                if self.start_year > other.start_year:
                    self_revenue = np.roll(self_revenue, (self.start_year - start_year))

            # If the length of other_revenue data is less than project duration
            if len(other_revenue) < end_year - start_year + 1:
                # Modify the size of other_revenue data
                other_revenue.resize(end_year - start_year + 1, refcheck=False)

                # If other.start_year > self.start_year, roll the array n steps to the right;
                # n = other.start_year - start_year
                if other.start_year > self.start_year:
                    other_revenue = np.roll(
                        other_revenue, (other.start_year - start_year)
                    )

            return self_revenue - other_revenue

        # If "other" is int, float, or numpy array
        elif isinstance(other, (int, float, np.ndarray)):
            return self.revenue() - other

        else:
            raise LiftingException(
                f"Must subtract an instance of Lifting with another instance "
                f"of Lifting or int/float "
                f"{other}({other.__class__.__qualname__}) is not an instance of "
                f"Lifting or int/float"
            )

    def __mul__(self, other):
        # Multiplication is allowed only with an integer/a float
        if isinstance(other, (int, float)):
            return Lifting(
                start_year=self.start_year,
                end_year=self.end_year,
                lifting_rate=self.lifting_rate * other,
                price=self.price,
                prod_year=self.prod_year,
                ghv=self.ghv,
                prod_rate=self.prod_rate,
                prod_rate_baseline=self.prod_rate_baseline,
            )

        else:
            raise LiftingException(
                f"Must multiply with an integer or a float; "
                f"{other} is not an integer nor a float"
            )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        # Between two instances of Lifting
        if isinstance(other, Lifting):
            return np.sum(self.revenue()) / np.sum(other.revenue())

        # Between an instance of Lifting and an instance of Tangible/Intangible/OPEX/ASR
        elif isinstance(other, (CapitalCost, Intangible, OPEX, ASR)):
            return np.sum(self.revenue()) / np.sum(other.expenditures())

        # Between an instance of Lifting and an integer/float
        elif isinstance(other, (int, float)):
            # Cannot divide with zero
            if other == 0:
                raise LiftingException(f"Cannot divide with zero")

            else:
                return Lifting(
                    start_year=self.start_year,
                    end_year=self.end_year,
                    lifting_rate=self.lifting_rate / other,
                    price=self.price,
                    prod_year=self.prod_year,
                    ghv=self.ghv,
                    prod_rate=self.prod_rate,
                    prod_rate_baseline=self.prod_rate_baseline,
                )

        elif isinstance(other, np.ndarray):
            return np.sum(self.revenue()) / np.sum(other)

        else:
            raise LiftingException(
                f"Does not allow division operation of an instance of Lifting "
                f"and {other.__class__.__qualname__}"
            )



""" Specify callable classes and methods from package 'econ' """

from .revenue import Lifting
from .costs import (
    GeneralCost,
    CapitalCost,
    Intangible,
    OPEX,
    ASR,
)
from .depreciation import (
    straight_line_depreciation_rate,
    straight_line_book_value,
    declining_balance_depreciation_rate,
    declining_balance_book_value,
    psc_declining_balance_depreciation_rate,
    psc_declining_balance_book_value,
    unit_of_production_rate,
    unit_of_production_book_value,
)
from .selection import (
    DeprMethod,
    FluidType,
    TaxType,
    TaxSplitTypeCR,
    YearReference,
    TaxRegime,
    TaxPaymentMode,
    FTPTaxRegime,
    GrossSplitRegime,
    DiscountingMode,
    ContractType,
    NPVSelection,
    OtherRevenue,
    InflationAppliedTo,
    OptimizationParameter,
    OptimizationTarget,
)
from .indicator import (
    pot,
    npv,
    xnpv,
    irr,
    xirr,
    npv_nominal_terms,
    npv_real_terms,
    npv_skk_nominal_terms,
    npv_skk_real_terms,
    npv_point_forward,
    pot_psc,
)



import numpy as np
from . import RPDModel


def oil_ltp_predict(volume: float) -> np.ndarray:
    """Oil forecast based on SKK Migas LTP Model
    ==============================================
    This function automatically set the RDPModel parameters with
    SKK Migas LTP forecast for oil. There might be discrepancy
    from the rounding errors with SKK Migas official LTP.

    Parameter
    ---
    volume : float
        Volume of resources (MSTB)

    Return
    ---
    rate : ndarray
        Yearly production rate (MSTB/yr)
    """
    if volume <= 4000:
        year_rampup = 0
        drate = 0.200633347
        q_plateau_ratio = 0.182937586835236
        q_min_ratio = 0.0014828
    elif volume <= 10_000:
        year_rampup = 1
        drate = 0.160502926
        q_plateau_ratio = 0.147021007393939
        q_min_ratio = 0.0036657
    elif volume <= 20_000:
        year_rampup = 1
        drate = 0.160549732
        q_plateau_ratio = 0.128715547194678
        q_min_ratio = 0.0037641
    elif volume <= 50_000:
        year_rampup = 2
        drate = 0.138588938
        q_plateau_ratio = 0.111027381539511
        q_min_ratio = 0.0060461
    elif volume <= 100_000:
        year_rampup = 2
        drate = 0.130770622
        q_plateau_ratio = 0.0972034569873284
        q_min_ratio = 0.0071092
    elif volume <= 200_000:
        year_rampup = 3
        drate = 0.131308644
        q_plateau_ratio = 0.0861313010167252
        q_min_ratio = 0.0081037
    elif volume <= 500_000:
        year_rampup = 3
        drate = 0.109315569
        q_plateau_ratio = 0.0742950873371555
        q_min_ratio = 0.0115846
    else:
        year_rampup = 3
        drate = 0.103678476
        q_plateau_ratio = 0.0650446693978552
        q_min_ratio = 0.0137343
    rpd = RPDModel(year_rampup, drate, q_plateau_ratio, q_min_ratio)
    return rpd.predict(volume)


def gas_ltp_predict(volume: float) -> np.ndarray:
    """Gas forecast based on SKK Migas LTP Model
    ==============================================
    This method automatically set the class parameters with
    SKK Migas LTP forecast for gas. There might be discrepancy
    from the rounding errors.

    Parameter
    ---
    volume : float
        Volume of resources (BSCF)

    Return
    ---
    rate : ndarray
        Yearly production rate (BSCF/yr)
    """
    if volume <= 100:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.123459112068285
        q_min_ratio = 0.0000125385020719761
    elif volume <= 200:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.0990140494482555
        q_min_ratio = 0.0000279329515875812
    elif volume <= 500:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.0900958146440382
        q_min_ratio = 0.0000423616992187692
    elif volume <= 1000:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.0763472955954202
        q_min_ratio = 0.0000997148666720888
    elif volume <= 2000:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.0662490413481761
        q_min_ratio = 0.000240349571639406
    else:
        year_rampup = 2
        drate = 0.510825623765991
        q_plateau_ratio = 0.0585312743031216
        q_min_ratio = 0.000589860388919471
    rpd = RPDModel(year_rampup, drate, q_plateau_ratio, q_min_ratio)
    return rpd.predict(volume)



"""
Specify callable classes/methods from package io
"""

from .aggregator import Aggregate
from .parse import InitiateContract
from .spreadsheet import Spreadsheet
from .config import (
    GeneralConfigData,
    FiscalConfigData,
    OilLiftingData,
    GasLiftingData,
    LPGPropaneLiftingData,
    LPGButaneLiftingData,
    SulfurLiftingData,
    ElectricityLiftingData,
    CO2LiftingData,
    CapitalCostData,
    IntangibleCostData,
    OPEXData,
    ASRCostData,
    PSCCostRecoveryData,
    PSCGrossSplitData,
    SensitivityData,
    MonteCarloData,
    OptimizationData,
)
from .plot_generator import get_uncertainty_plot
